/******************************************************************************

  $Header: f:\\RCS\\F\\devel\\pilot\\cppconduit\\LinkSyncConduitInstaller\\main.cpp,v 1.12 2002-09-06 16:43:42-04 eshayne Exp eshayne $
  $Revision: 1.12 $
 

  $Log: main.cpp,v $
  Revision 1.12  2002-09-06 16:43:42-04  eshayne
  updated version number

  Revision 1.11  2002-08-30 21:48:52-04  eshayne
  updated version number to 1.2

  Revision 1.10  2002-08-10 20:37:16-04  eshayne
  changed name to Link Synchronizer

  Revision 1.9  2002-07-21 10:24:59-04  eshayne
  added comment and license blocks


	----------------------------------------
    Link Synchronizer
    Copyright (C)2002 Ethan Shayne

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include <iostream.h>
#include <windows.h>
#include <string>
#include "LinkSyncConduitInstallerException.h"
#include <assert.h>
#include <CondMgr.h>
#include <HSAPI.h>

#define LINKSYNC_VERSION_NUMBER "1.2.1"


// this is the function signature for CmGetHotSyncExecPath()
typedef int (WINAPI* CmGetPathPtr)(TCHAR*, int*);
typedef int (WINAPI* CmInstallPtr)(const char*, int);  // CmInstallCreator()
typedef int (WINAPI* CmSetNamePtr)(const char*, const TCHAR*); // CmSetCreatorName()
typedef int (WINAPI* CmSetTitlePtr)(const char*, const TCHAR*); // CmSetCreatorTitle()
typedef int (WINAPI* CmSetDirectoryPtr)(const char*, const TCHAR*); // CmSetCreatorDirectory()
typedef int (WINAPI* CmSetFilePtr)(const char*, const TCHAR*); // CmSetCreatorFile();
typedef int (WINAPI* CmSetPriorityPtr)(const char*, DWORD); // CmSetCreatorPriority();
typedef int (WINAPI* CmRemoveConduitPtr)(const char*); // CmRemoveConduitByCreatorID();
typedef long (WINAPI* HsSetAppStatusPtr)(HsStatusType, DWORD); // HsSetAppStatus();


HINSTANCE LoadDLL(const stlport::string& strDir, const stlport::string& strFilename)
{
	HINSTANCE hDLL = NULL;
	stlport::string strFullFilePath = strDir + "\\" + strFilename;
	hDLL = LoadLibrary(strFullFilePath.c_str());
	if (!hDLL)
	{
		DWORD dwErr = GetLastError();
		hDLL = NULL;
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not load DLL " + strFullFilePath, dwErr);
	}

	return hDLL;
}


void UnloadDLL(HINSTANCE& hDLL, const stlport::string& strDLLName)
{
	BOOL bSuccess = FreeLibrary(hDLL);
	if (bSuccess != TRUE)
	{
		DWORD dwErr = GetLastError();
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "FreeLibrary() failed for " + strDLLName, dwErr);
	}
	hDLL = NULL;
}


FARPROC FindFunctionPtr(HINSTANCE hDLL, const stlport::string& strFunctionName)
{
	FARPROC pfn = GetProcAddress(hDLL, strFunctionName.c_str());
	if (!pfn)
	{
		DWORD dwErr = GetLastError();
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not find function " + strFunctionName, dwErr);
	}

	return pfn;
}


stlport::string FindDLLDirectory(const stlport::string& strTempInstallDir)
{
	// load our temporary version of CondMgr.dll
	char acRealPath[2048];
	int iRealPathSize = 2048;
	HINSTANCE hTempCondMgrDLL = NULL;
	try {
		hTempCondMgrDLL = LoadDLL(strTempInstallDir, "CondMgr.dll");

		// get a handle to the CmGetHotSyncExecPath() function within our temporary
		// version of CondMgr.dll
		CmGetPathPtr pfnGetPath = (CmGetPathPtr) FindFunctionPtr(hTempCondMgrDLL, "CmGetHotSyncExecPath");

		// execute the function to get the real DLL path
		int iErr = (*pfnGetPath)(acRealPath, &iRealPathSize);
		if (iErr != 0)
			throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "CmGetHotSyncExecPath() call failed", iErr);
	}
	catch (...)
	{
		if (hTempCondMgrDLL)
			UnloadDLL(hTempCondMgrDLL, "CondMgr.dll");
		throw;
	}

	// free our handle to the temporary DLL
	if (hTempCondMgrDLL)
		UnloadDLL(hTempCondMgrDLL, "CondMgr.dll");

	stlport::string strHotSyncPath(acRealPath, iRealPathSize);

	// strip out the HotSync.exe to get the actual path
	stlport::string strRealDLLPath;
	size_t uiLastSlashPos = strHotSyncPath.find_last_of("\\");
	if (uiLastSlashPos == stlport::string::npos)
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not parse hotsync path " + strHotSyncPath, 0);
	strRealDLLPath = strHotSyncPath.substr(0, uiLastSlashPos);

	return strRealDLLPath;
}


stlport::string GetConduitDirectory()
{
	// for now, assumes the current directory (.) - but we need
	// to get the full path to pass to the Conduit Manager

	char acPath[2048];
	DWORD dwBufferSize = 2048;
	DWORD dwActualPathSize = GetCurrentDirectory(dwBufferSize, acPath);
	DWORD dwErr = GetLastError();
	if (dwActualPathSize == 0 || dwActualPathSize > 2047)
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not determine current directory path", dwErr);

	return stlport::string(acPath);
}


void InstallConduit(const char* pszCreator, HINSTANCE hCondMgr, const char* pszConduitPathOverride)
{
	assert(hCondMgr);
	assert(pszCreator);

	CmInstallPtr pfnInstall = (CmInstallPtr)FindFunctionPtr(hCondMgr, "CmInstallCreator");
	int err = (*pfnInstall)(pszCreator, CONDUIT_APPLICATION);
	if (err != 0)
	{
		if (err == -1005)
		{
			cout << "LinkSync conduit already installed... Upgrading..." << endl;
		}
		else
			throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not install conduit", err);
	}

	stlport::string strConduitFullPath;
	if (pszConduitPathOverride == NULL)
		strConduitFullPath = GetConduitDirectory() + "\\LinkSyncConduit.dll";
	else
		strConduitFullPath = stlport::string(pszConduitPathOverride);
	cout << "Using conduit path " << strConduitFullPath << "..." << endl;
	CmSetNamePtr pfnSetName = (CmSetNamePtr)FindFunctionPtr(hCondMgr, "CmSetCreatorName");
	err = (*pfnSetName)(pszCreator, strConduitFullPath.c_str());
	if (err != 0)
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not set conduit DLL", err);

	CmSetTitlePtr pfnSetTitle = (CmSetTitlePtr)FindFunctionPtr(hCondMgr, "CmSetCreatorTitle");
	err = (*pfnSetTitle)(pszCreator, "Link Synchronizer");
	if (err != 0)
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not set conduit title", err);

	CmSetDirectoryPtr pfnSetDirectory = (CmSetDirectoryPtr)FindFunctionPtr(hCondMgr, "CmSetCreatorDirectory");
	err = (*pfnSetDirectory)(pszCreator, "LinkSync");
	if (err != 0)
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not set conduit directory", err);

	CmSetFilePtr pfnSetFile = (CmSetFilePtr)FindFunctionPtr(hCondMgr, "CmSetCreatorFile");
	err = (*pfnSetFile)(pszCreator, "LinkSync");
	if (err != 0)
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not set conduit file", err);

	CmSetPriorityPtr pfnSetPriority = (CmSetPriorityPtr)FindFunctionPtr(hCondMgr, "CmSetCreatorPriority");
	err = (*pfnSetPriority)(pszCreator, 2);
	if (err != 0)
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not set conduit priority", err);
}


void UninstallConduit(const char* pszCreator, HINSTANCE hCondMgr)
{
	assert(hCondMgr);
	assert(pszCreator);

	CmRemoveConduitPtr pfnRemoveConduit = (CmRemoveConduitPtr)FindFunctionPtr(hCondMgr, "CmRemoveConduitByCreatorID");
	int iNumRemoved = (*pfnRemoveConduit)(pszCreator);
	if (iNumRemoved < 0)
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not uninstall conduit", iNumRemoved);
}


void RestartHotSync(HINSTANCE hHSAPI)
{
	assert(hHSAPI);

	HsSetAppStatusPtr pfnSetStatus = (HsSetAppStatusPtr)FindFunctionPtr(hHSAPI, "HsSetAppStatus");
	long err = (*pfnSetStatus)(HsRestart, HSFLAG_NONE);
	if (err != 0)
		throw LinkSyncConduitInstallerException(__FILE__, __LINE__, "Could not restart HotSync", err);
}


void LicenseSummary()
{
	cout << "Link Synchronizer comes with ABSOLUTELY NO WARRANTY; for details please see" << endl
		 << "http://www.eshayne.com/LinkSync/license.txt" << endl
		 << "This is free software, and you are welcome to redistribute it" << endl
		 << "under certain conditions; please see the GNU General Public License for" << endl
		 << "details." << endl
		 << endl
		 << "You should have received a copy of the GNU General Public License (which covers" << endl
		 << "all warranty and licensing issues for this product) with the Link Synchronizer utility" << endl
		 << "in a file named license.txt. If you did not, it can be obtained from the" << endl
		 << "product's web site (http://www.eshayne.com/LinkSync), or you can write to the" << endl
		 << "Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA" << endl;
}


void Version()
{
	cout << "LinkSync v" << LINKSYNC_VERSION_NUMBER << ", Copyright (C) 2002 Ethan Shayne" << endl
		 << "http://www.eshayne.com/LinkSync" << endl
		 << endl;
	LicenseSummary();
}


void Help()
{
	cout << "LinkSyncConduitInstaller /install" << endl
		 << "LinkSyncConduitInstaller /uninstall" << endl
		 << "LinkSyncConduitInstaller /version" << endl
		 << "LinkSyncConduitInstaller /license" << endl
		 << "(Please see http://www.eshayne.com/LinkSync for full documentation)" << endl
		 << endl;
	Version();
}


int main(int argc, char** argv)
{
	// are we installing or uninstalling?
	// if they don't specify, assume install
	char cAction = 'i';
	if (argc >= 2)
	{
		char* pszFlag = argv[1];
		if (strcmp(pszFlag, "/install") == 0)
			cAction = 'i';
		else if (strcmp(pszFlag, "/uninstall") == 0)
			cAction = 'u';
		else if (strcmp(pszFlag, "/version") == 0)
		{
			Version();
			exit(0);
		}
		else if (strcmp(pszFlag, "/license") == 0)
		{
			LicenseSummary();
			exit(0);
		}
		else
		{
			Help();
			exit(1);
		}
	}

	if (cAction == 'i')
		cout << "Installing Link Synchronizer conduit..." << endl;
	else
		cout << "Uninstalling Link Synchronizer conduit..." << endl;

	const char* pszCreator = "LSnc";

	HINSTANCE hCondMgr = NULL;
	HINSTANCE hHSAPI = NULL;
	bool bSuccess = false;

	try {
		// find the real directory where the real DLLs are stored
		// use the current directory (.) as the assumed place where our
		// temporary version of CondMgr.dll resides
		stlport::string strRealDLLDir = FindDLLDirectory(".");

		// load the real CondMgr.dll
		hCondMgr = LoadDLL(strRealDLLDir, "CondMgr.dll");

		if (cAction == 'i')
		{
			// install the conduit
			try {
				char* pszConduitPathOverride = NULL;
				if (argc >= 3)
					pszConduitPathOverride = argv[2];
				InstallConduit(pszCreator, hCondMgr, pszConduitPathOverride);
				bSuccess = true;
			}
			catch (const LinkSyncConduitInstallerException& e2)
			{
				cerr << "Error installing conduit: " << e2 << endl;
			}
			catch (...)
			{
				cerr << "Unknown error installing conduit" << endl;
			}

			if (!bSuccess)
			{
				cerr << "Backing out installation..." << endl;
				try {
					UninstallConduit(pszCreator, hCondMgr);
					cerr << "Link Synchronizer conduit installation successfully backed out." << endl;
				}
				catch (const LinkSyncConduitInstallerException& e2)
				{
					cerr << "Could not back out Link Synchronizer conduit installation: " << e2 << endl;
				}
			}
		}
		else
		{
			// uninstall the conduit
			UninstallConduit(pszCreator, hCondMgr);
			bSuccess = true;
		}

		// unload CondMgr.dll
		UnloadDLL(hCondMgr, "CondMgr.dll");

		if (bSuccess)
		{
			// load the real HSAPI.dll
			hHSAPI = LoadDLL(strRealDLLDir, "HSAPI.dll");

			// bounce HotSync.exe
			cout << "Restarting HotSync..." << endl;
			RestartHotSync(hHSAPI);
			cout << "HotSync successfully restarted." << endl;

			UnloadDLL(hHSAPI, "HSAPI.dll");
		}
	}
	catch (const LinkSyncConduitInstallerException& e)
	{
		cerr << e << endl;
	}
	catch (...)
	{
		cerr << "caught unknown exception" << endl;
	}

	if (hCondMgr)
	{
		try {
			UnloadDLL(hCondMgr, "CondMgr.dll");
		}
		catch (const LinkSyncConduitInstallerException& e)
		{
			cerr << "Error encountered while unloading CondMgr.dll: " << e << endl;
		}
		catch (...)
		{
			cerr << "caught unknown exception while unloading CondMgr.dll" << endl;
		}
	}

	if (hHSAPI)
	{
		try {
			UnloadDLL(hHSAPI, "HSAPI.dll");
		}
		catch (const LinkSyncConduitInstallerException& e)
		{
			cerr << "Error encountered while unloading HSAPI.dll: " << e << endl;
		}
		catch (...)
		{
			cerr << "caught unknown exception while unloading HSAPI.dll" << endl;
		}
	}

	if (bSuccess)
	{
		cout << "Link Synchronizer conduit ";
		if (cAction == 'i')
			cout << "installation";
		else
			cout << "uninstallation";
		cout << " successfully completed!" << endl;
		return 0;
	}
	else
	{
		cout << "Link Synchronizer conduit ";
		if (cAction == 'i')
			cout << "installation";
		else
			cout << "uninstallation";
		cout << " failed." << endl;
		return 1;
	}
}
