/******************************************************************************

  $Header: f:\\RCS\\F\\devel\\pilot\\cppconduit\\LinkSyncConduit\\LinkSyncDesktopFavorite.cpp,v 1.12 2002-09-06 16:43:16-04 eshayne Exp eshayne $
  $Revision: 1.12 $
 

  $Log: LinkSyncDesktopFavorite.cpp,v $
  Revision 1.12  2002-09-06 16:43:16-04  eshayne
  create subfolders if necessary

  Revision 1.11  2002-08-30 21:45:28-04  eshayne
  - added subfolder support
  - fixed bug in LinkSyncDesktopFavorite::GetAllFavorites(): was leaving file handles open

  Revision 1.10  2002-08-10 20:34:32-04  eshayne
  changed name to Link Synchronizer

  Revision 1.9  2002-07-21 10:24:17-04  eshayne
  added comment and license blocks


	----------------------------------------
    Link Synchronizer
    Copyright (C)2002 Ethan Shayne

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include <string.h>
#include "LinkSyncDesktopFavorite.h"
#include "LinkSyncConduitException.h"
#include <string>
#include <windows.h>
#include <HSLog.h>
#include <assert.h>
#include <SyncMgr.h>
#include <direct.h>
#include <errno.h>

#define LINKSYNC_VERSION_NUMBER "1.2.1"

stlport::string* LinkSyncDesktopFavorite::mgpstrUserFavoritesRootDirectory = NULL;


LinkSyncDesktopFavorite::LinkSyncDesktopFavorite()
: mpstrDirectory(NULL),
  mpdwHHRecordId(NULL),
  mpPalmDBFormattedData(NULL)
{
}


LinkSyncDesktopFavorite::LinkSyncDesktopFavorite(const LinkSyncDesktopFavorite& copy)
: mstrUrl(copy.mstrUrl),
  mstrName(copy.mstrName),
  mpstrDirectory(NULL),
  mpdwHHRecordId(NULL),
  mpPalmDBFormattedData(NULL)
{
	if (copy.mpstrDirectory)
		SetDirectory(copy.GetDirectory());
	if (copy.mpdwHHRecordId)
		SetHHRecordId(copy.GetHHRecordId());
	if (copy.mpPalmDBFormattedData)
		SetPalmDBFormattedData(copy.mpPalmDBFormattedData->GetData(), copy.mpPalmDBFormattedData->GetSize());
}


LinkSyncDesktopFavorite& LinkSyncDesktopFavorite::operator=(const LinkSyncDesktopFavorite& copy)
{
	if (this == &copy)
		return *this;

	mstrUrl = copy.mstrUrl;
	mstrName = copy.mstrName;
	if (copy.mpstrDirectory)
		SetDirectory(copy.GetDirectory());
	if (copy.mpdwHHRecordId)
		SetHHRecordId(copy.GetHHRecordId());
	if (copy.mpPalmDBFormattedData)
		SetPalmDBFormattedData(copy.mpPalmDBFormattedData->GetData(), copy.mpPalmDBFormattedData->GetSize());

	return *this;
}

	
LinkSyncDesktopFavorite::LinkSyncDesktopFavorite(const char* pszName,
												 const char* pszUrl)
: mpstrDirectory(NULL),
  mpdwHHRecordId(NULL),
  mpPalmDBFormattedData(NULL)
{
	SetName(pszName);
	SetUrl(pszUrl);
}


LinkSyncDesktopFavorite::LinkSyncDesktopFavorite(const char* pszPalmDBBuffer)
: mpstrDirectory(NULL),
  mpdwHHRecordId(NULL),
  mpPalmDBFormattedData(NULL)
{
	SetFromPalmDBBuffer(pszPalmDBBuffer);
}


LinkSyncDesktopFavorite::LinkSyncDesktopFavorite(const LinkSyncConduitRawData& rawData)
: mpstrDirectory(NULL),
  mpdwHHRecordId(NULL),
  mpPalmDBFormattedData(NULL)
{
	SetFromPalmDBBuffer((const char*)(rawData.GetData()));
}

	
LinkSyncDesktopFavorite::~LinkSyncDesktopFavorite()
{
	if (mpstrDirectory)
		delete mpstrDirectory;
	if (mpdwHHRecordId)
		delete mpdwHHRecordId;
	if (mpPalmDBFormattedData)
		delete mpPalmDBFormattedData;
}


const char* LinkSyncDesktopFavorite::GetName() const
{
	return mstrName.c_str();
}


const char* LinkSyncDesktopFavorite::GetUrl() const
{
	return mstrUrl.c_str();
}


const char* LinkSyncDesktopFavorite::GetDirectory() const
{
	if (!mpstrDirectory)
		throw LinkSyncConduitException(__FILE__, __LINE__, "LinkSyncDesktopFavorite::GetDirectory(): no directory defined for this record");

	return mpstrDirectory->c_str();
}


DWORD LinkSyncDesktopFavorite::GetHHRecordId() const
{
	if (!mpdwHHRecordId)
		throw LinkSyncConduitException(__FILE__, __LINE__, "LinkSyncDesktopFavorite::GetHHRecordId(): no handheld record id defined for this record");

	return *mpdwHHRecordId;
}


void LinkSyncDesktopFavorite::SetName(const char* pszName)
{
	mstrName = pszName;
	if (mpPalmDBFormattedData)
	{
		delete mpPalmDBFormattedData;
		mpPalmDBFormattedData = NULL;
	}
}


void LinkSyncDesktopFavorite::SetUrl(const char* pszUrl)
{
	mstrUrl = pszUrl;
	if (mpPalmDBFormattedData)
	{
		delete mpPalmDBFormattedData;
		mpPalmDBFormattedData = NULL;
	}
}


void LinkSyncDesktopFavorite::SetDirectory(const char* pszDirectory)
{
	if (mpstrDirectory)
		delete mpstrDirectory;

	mpstrDirectory = new stlport::string(pszDirectory);
}


void LinkSyncDesktopFavorite::SetHHRecordId(const DWORD dwRecordId)
{
	if (mpdwHHRecordId)
		delete mpdwHHRecordId;

	mpdwHHRecordId = new DWORD(dwRecordId);
}


void LinkSyncDesktopFavorite::ReadFromFile(const char* pszFilename, const char* pszDirectory)
{
	if (!pszFilename)
		throw LinkSyncConduitException(__FILE__, __LINE__, "LinkSyncDesktopFavorite::ReadFromFile(): filename supplied as NULL");

	// remove the trailing ".url" to get the favorite "name"
	stlport::string strFilename = pszFilename;
	stlport::string strName;
	if (strFilename.length() > 4)
		strName = strFilename.substr(0, (strFilename.length() - 4));

	if (strName.length() == 0)
	{
		stlport::string strErr = "LinkSyncDesktopFavorite::ReadFromFile(): could not parse out name from filename '" + strFilename + "'";
		throw LinkSyncConduitException(__FILE__, __LINE__, strErr.c_str());
	}

	SetName(strName.c_str());
	stlport::string strDirectory;
	if (pszDirectory)
		strDirectory = stlport::string(pszDirectory);
	else
		strDirectory = LinkSyncDesktopFavorite::GetUserFavoritesRootDirectory();
	SetDirectory(strDirectory.c_str());

	// first, load the contents of the file into memory
	stlport::string strFileContents;
	{
		ifstream fDesktopFavoriteFile(GetFilename().c_str(), ios::in|ios::nocreate);
		if (!fDesktopFavoriteFile)
		{
			stlport::string strLogEntry("LinkSyncDesktopFavorite::ReadFromFile(): could not open desktop file '");
			strLogEntry += GetFilename();
			strLogEntry += "' for reading";
			throw LinkSyncConduitException(__FILE__, __LINE__, strLogEntry.c_str());
		}

		char ch;
		while (fDesktopFavoriteFile.get(ch))
			strFileContents += ch;
	}

	// parse the file contents to find the URL
	// the URL is right after the "URL=", which follows "[InternetShortcut]"
	size_t uiShortcutLabelPos = strFileContents.find("[InternetShortcut]");
	if (uiShortcutLabelPos == stlport::string::npos)
	{
		stlport::string strLogEntry("LinkSyncDesktopFavorite::ReadFromFile(): could not find [InternetShortcut] in file " + GetFilename() + " contents: '");
		strLogEntry += strFileContents;
		strLogEntry += "'";
		throw LinkSyncConduitException(__FILE__, __LINE__, strLogEntry.c_str());
	}
	size_t uiBaseUrlLabelPos = strFileContents.find("URL=", uiShortcutLabelPos + 18);
	if (uiBaseUrlLabelPos == stlport::string::npos)
	{
		stlport::string strLogEntry("LinkSyncDesktopFavorite::ReadFromFile(): could not find URL= in file " + GetFilename() + " contents: '");
		strLogEntry += strFileContents;
		strLogEntry += "'";
		throw LinkSyncConduitException(__FILE__, __LINE__, strLogEntry.c_str());
	}


	size_t uiUrlStartPos = uiBaseUrlLabelPos + 4;
	size_t uiUrlEndPos = strFileContents.find("\n", uiUrlStartPos);
	if (uiUrlEndPos == stlport::string::npos)
	{
		stlport::string strLogEntry("LinkSyncDesktopFavorite::ReadFromFile(): could not find closing newline in file " + GetFilename() + " contents: '");
		strLogEntry += strFileContents;
		strLogEntry += "'";
		throw LinkSyncConduitException(__FILE__, __LINE__, strLogEntry.c_str());
	}

	stlport::string strUrl = strFileContents.substr(uiUrlStartPos, (uiUrlEndPos - uiUrlStartPos));
	SetUrl(strUrl.c_str());
}

void LinkSyncDesktopFavorite::SplitDirectory(const stlport::string& strDirectory, stlport::vector<stlport::string>& colDirectoryPieces) const
{
	size_t uiSlashPos = strDirectory.find("\\");
	if (uiSlashPos == stlport::string::npos)
	{
		colDirectoryPieces.push_back(strDirectory);
		return;
	}

	stlport::string strDirectoryPiece = strDirectory.substr(0, uiSlashPos);
	colDirectoryPieces.push_back(strDirectoryPiece);
	stlport::string strLeftoverDirectory = strDirectory.substr(uiSlashPos + 1);
	SplitDirectory(strLeftoverDirectory, colDirectoryPieces);
}


void LinkSyncDesktopFavorite::CreateDirectoryForFile(const stlport::string& strFullFilename) const
{
	// parse out the directory from the filename
	size_t uiLastSlashPos = strFullFilename.find_last_of("\\");
	if (uiLastSlashPos == stlport::string::npos)
	{
		stlport::string strLogEntry("LinkSyncDesktopFavorite::CreateDirectoryForFile(\"" + strFullFilename + "\"): could not find last \\");
		throw LinkSyncConduitException(__FILE__, __LINE__, strLogEntry.c_str());
	}
	stlport::string strDirectory = strFullFilename.substr(0, uiLastSlashPos);

	stlport::vector<stlport::string> colDirectoryPieces;
	SplitDirectory(strDirectory, colDirectoryPieces);

	stlport::vector<stlport::string>::const_iterator iterPieces;
	stlport::string strDirectoryToCreate;
	for (iterPieces = colDirectoryPieces.begin(); iterPieces != colDirectoryPieces.end(); iterPieces++)
	{
		const stlport::string& strDirectoryPiece = *iterPieces;
		if (!strDirectoryToCreate.empty())
			strDirectoryToCreate += "\\";
		strDirectoryToCreate += strDirectoryPiece;
		int iRetVal = _mkdir(strDirectoryToCreate.c_str());
		if (iRetVal == -1 && errno != EEXIST)
		{
			stlport::string strLogEntry("LinkSyncDesktopFavorite::CreateDirectoryForFile(\"" + strFullFilename + "\"): could not create directory \"" + strDirectoryToCreate + "\"");
			throw LinkSyncConduitException(__FILE__, __LINE__, strLogEntry.c_str(), errno);
		}

		if (iRetVal == 0)
			LogAddFormattedEntry(slText, true, "LinkSynchronizer created directory \"%s\"", strDirectoryToCreate.c_str());
	}
}


void LinkSyncDesktopFavorite::WriteToFile() const
{
	// this function is written very awkwardly - unfortunately, it's the only
	// way I could get this to work under VC++ 6.0

	stlport::string strFullFilename = GetFilename();
	ofstream fDesktopFavoriteFile(strFullFilename.c_str(), iostream::out);
	if (fDesktopFavoriteFile)
	{
		fDesktopFavoriteFile << "[DEFAULT]" << endl
							 << "BASEURL=" << GetUrl() << endl
							 << "[InternetShortcut]" << endl
							 << "URL=" << GetUrl() << endl;
	}
	else
	{
		DWORD dwErr = GetLastError();
		if (dwErr != ERROR_PATH_NOT_FOUND)
		{
			stlport::string strLogEntry("LinkSyncDesktopFavorite::WriteToFile(): could not open desktop file '");
			strLogEntry += strFullFilename;
			strLogEntry += "' for writing";
			throw LinkSyncConduitException(__FILE__, __LINE__, strLogEntry.c_str(), dwErr);
		}
		else
		{
			// the directory specified does not exist - create it and try again
			CreateDirectoryForFile(strFullFilename);
			ofstream fDesktopFavoriteFile2(strFullFilename.c_str(), iostream::out);
			if (fDesktopFavoriteFile2)
			{
				fDesktopFavoriteFile2 << "[DEFAULT]" << endl
									  << "BASEURL=" << GetUrl() << endl
									  << "[InternetShortcut]" << endl
									  << "URL=" << GetUrl() << endl;
			}
			else
			{
				dwErr = GetLastError();
				stlport::string strLogEntry("LinkSyncDesktopFavorite::WriteToFile(): could not open desktop file '");
				strLogEntry += strFullFilename;
				strLogEntry += "' for writing";
				throw LinkSyncConduitException(__FILE__, __LINE__, strLogEntry.c_str(), dwErr);
			}
		}
	}
}


LinkSyncConduitRawData LinkSyncDesktopFavorite::GetPalmDBFormattedData()
{
	if (!mpPalmDBFormattedData)
	{
		mpPalmDBFormattedData = new LinkSyncConduitRawData();
		assert(mpPalmDBFormattedData);
	
		mpPalmDBFormattedData->SetData(GetName(), strlen(GetName()) + 1);
		// no need to tell AddData() to null-delimit - there's already a null there
		mpPalmDBFormattedData->AddData(GetUrl(), strlen(GetUrl()) + 1, false);
	}

	assert(mpPalmDBFormattedData);
	return *mpPalmDBFormattedData;
}


void LinkSyncDesktopFavorite::GetAllFavorites(FavoriteList& colFavoritesList,
											  const char* pszPath,
											  const char* pszDiffFromRoot)
{
	stlport::string strRootPath;
	if (pszPath)
		strRootPath = stlport::string(pszPath);
	else
		strRootPath = LinkSyncDesktopFavorite::GetUserFavoritesRootDirectory();

	stlport::string strSearchSpec = strRootPath + "\\" + "*";
	WIN32_FIND_DATA fileData;
	HANDLE hFile = NULL;
	stlport::vector<stlport::string> colSubdirs;
	unsigned long ulYieldCount = 0;
	long err = 0;
	try {
		hFile = FindFirstFile(strSearchSpec.c_str(), &fileData);
		BOOL bFoundFile = (hFile != INVALID_HANDLE_VALUE);
		while (bFoundFile)
		{
			if (++ulYieldCount >= 100)
			{
				ulYieldCount = 0;
				err = SyncYieldCycles(1);
				if (err != 0)
				{
					LogAddFormattedEntry(slWarning, true, "Could not yeild cycles during sync while loading all favorites from desktop, err = %d", err);
				}
			}

			if ((fileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
			{
				// if it's not a subdirectory, check that it's a url/bookmark/favorite
				int iFilenameLen = strlen(fileData.cFileName);
				int iOffset = iFilenameLen - 4;
				char* pszLast4 = fileData.cFileName;
				if (iOffset >= 0)
					pszLast4 = fileData.cFileName + iOffset;
				if (strcmp(pszLast4, ".url") == 0)
				{
					// we got one!
					LinkSyncDesktopFavorite fav;
					stlport::string strFilename;
					if (pszDiffFromRoot)
						strFilename = stlport::string(pszDiffFromRoot) + "\\";
					strFilename += stlport::string(fileData.cFileName);

					fav.ReadFromFile(strFilename.c_str());
					colFavoritesList.push_back(fav);
				}
			}
			else
			{
				// if it is a subdirectory, make sure it's not . or ..
				if (!(strcmp(fileData.cFileName, ".") == 0 || strcmp(fileData.cFileName, "..") == 0))
				{
					// if it's a valid subdirectory, add it to the list for processing after this directory
					// is done. we don't want to process it right away because we don't want to
					// keep too many file handles open
					colSubdirs.push_back(stlport::string(fileData.cFileName));
				}
			}
			bFoundFile = FindNextFile(hFile, &fileData);
		}
		FindClose(hFile);
		hFile = NULL;
	}
	catch (...)
	{
		if (hFile)
			FindClose(hFile);
		throw;
	}

	// process any subdirectories too
	stlport::vector<stlport::string>::const_iterator iterSubdirs;
	ulYieldCount = 0;

	for (iterSubdirs = colSubdirs.begin(); iterSubdirs != colSubdirs.end(); iterSubdirs++)
	{
		if (++ulYieldCount >= 100)
		{
			ulYieldCount = 0;
			err = SyncYieldCycles(1);
			if (err != 0)
			{
				LogAddFormattedEntry(slWarning, true, "Could not yeild cycles during sync while loading all favorites from desktop, err = %d", err);
			}
		}

		const stlport::string& strSubdir = (*iterSubdirs);

		stlport::string strSubRoot = strRootPath + "\\" + strSubdir;

		stlport::string strNewDiffFromRoot;
		if (pszDiffFromRoot)
			strNewDiffFromRoot += stlport::string(pszDiffFromRoot) + "\\";
		strNewDiffFromRoot += strSubdir;

		LinkSyncDesktopFavorite::GetAllFavorites(colFavoritesList, strSubRoot.c_str(), strNewDiffFromRoot.c_str());
	}
}


void LinkSyncDesktopFavorite::PutFavoritesIntoMap(const FavoriteList& colFavoriteList, FavoriteMap& colFavoriteMap)
{
	colFavoriteMap.erase(colFavoriteMap.begin(), colFavoriteMap.end());
	FavoriteList::const_iterator iterFav;
	unsigned long ulYieldCount = 0;
	long err = 0;
	for (iterFav = colFavoriteList.begin(); iterFav != colFavoriteList.end(); iterFav++)
	{
		if (++ulYieldCount >= 1000)
		{
			ulYieldCount = 0;
			err = SyncYieldCycles(1);
			if (err != 0)
			{
				LogAddFormattedEntry(slWarning, true, "Could not yeild cycles during sync while putting desktop favorite instances into a map structure, err = %d", err);
			}
		}

		colFavoriteMap[(*iterFav).GetName()] = (*iterFav);
	}
}


stlport::string LinkSyncDesktopFavorite::GetFilename() const
{
	return stlport::string(GetDirectory()) + "\\" + stlport::string(GetName()) + ".url";
}


DWORD LinkSyncDesktopFavorite::DeleteDesktopFile() const
{
	if (DeleteFile(GetFilename().c_str()) == 0)
		return GetLastError();
	else
		return 0;
}


bool LinkSyncDesktopFavorite::operator== (const LinkSyncDesktopFavorite& other)
{
	// "equal" in this case means are they effectively the same - do they have
	// the same name and URL, which is really all that matters
	return ((GetName() == other.GetName()) &&
			(GetUrl() == other.GetUrl()));
}


bool LinkSyncDesktopFavorite::operator< (const LinkSyncDesktopFavorite& other)
{
	return (GetName() < other.GetName());
}


bool LinkSyncDesktopFavorite::operator> (const LinkSyncDesktopFavorite& other)
{
	return (GetName() > other.GetName());
}


void LinkSyncDesktopFavorite::LoadUserFavoritesRootDirectoryFromRegistry()
{
	// find the current user's favorites directory by looking in the Windows
	// registry

	if (LinkSyncDesktopFavorite::mgpstrUserFavoritesRootDirectory)
		delete LinkSyncDesktopFavorite::mgpstrUserFavoritesRootDirectory;

	bool bRegistryKeyOpen = false;
	HKEY hKey;
	long lResult = ERROR_SUCCESS;
	BYTE baValue[5121];
	DWORD dwValueSize = 5121;
	try {
		lResult = RegOpenKeyEx(HKEY_CURRENT_USER,
									"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders",
									0,
									KEY_READ,
									&hKey);
		if (lResult != ERROR_SUCCESS)
			throw LinkSyncConduitException(__FILE__, __LINE__, "could not open registry key to determine user's favorites directory", lResult);

		DWORD dwValueType = 0;
		lResult = RegQueryValueEx(hKey,
								  "Favorites",
								  0,
								  &dwValueType,
								  baValue,
								  &dwValueSize);
		if (lResult != ERROR_SUCCESS || dwValueType != REG_SZ)
			throw LinkSyncConduitException(__FILE__, __LINE__, "could not read registry value to determine user's favorites directory", lResult);
	}
	catch (const LinkSyncConduitException&)
	{
		RegCloseKey(hKey);
		throw;
	}
	catch (...)
	{
		RegCloseKey(hKey);
		throw;
	}

	lResult = RegCloseKey(hKey);
	if (lResult != ERROR_SUCCESS)
		throw LinkSyncConduitException(__FILE__, __LINE__, "could not close registry key while determining user's favorites directory", lResult);

	// dwValueSize includes the null terminator, which we don't need/want for this stlport::string constructor
	stlport::string strRootDir = stlport::string((char*)baValue, dwValueSize - 1);
	stlport::string strLogEntry("LinkSynchronizer version ");
	strLogEntry += stlport::string(LINKSYNC_VERSION_NUMBER);
	strLogEntry += ": user's root favorites directory set to ";
	strLogEntry += strRootDir;
	LogAddFormattedEntry(slText, true, strLogEntry.c_str());
	LinkSyncDesktopFavorite::mgpstrUserFavoritesRootDirectory = new stlport::string(strRootDir);
}


const char* LinkSyncDesktopFavorite::GetUserFavoritesRootDirectory()
{
	if (!LinkSyncDesktopFavorite::mgpstrUserFavoritesRootDirectory)
	{
		LinkSyncDesktopFavorite::LoadUserFavoritesRootDirectoryFromRegistry();
		if (!LinkSyncDesktopFavorite::mgpstrUserFavoritesRootDirectory)
			throw LinkSyncConduitException(__FILE__, __LINE__, "could not access user's favorites directory");
	}

	return (*LinkSyncDesktopFavorite::mgpstrUserFavoritesRootDirectory).c_str();
}


void LinkSyncDesktopFavorite::SetPalmDBFormattedData(const void* pData, unsigned short usSize)
{
	if (mpPalmDBFormattedData)
	{
		delete mpPalmDBFormattedData;
		mpPalmDBFormattedData = NULL;
	}

	mpPalmDBFormattedData = new LinkSyncConduitRawData(pData, usSize);
}


void LinkSyncDesktopFavorite::SetFromPalmDBBuffer(const char* pszPalmDBBuffer)
{
	if (!pszPalmDBBuffer)
		return;

	unsigned short usNameLen = strlen(pszPalmDBBuffer);
	SetName(pszPalmDBBuffer);
	const char* pszUrlPosInBuffer = pszPalmDBBuffer + usNameLen + 1;
	SetUrl(pszUrlPosInBuffer);
}

