/******************************************************************************

  $Header: f:\\RCS\\f\\devel\\pilot\\cppconduit\\linksyncconduit\\LinkSyncDesktopLastState.cpp,v 1.6 2002-08-10 20:34:32-04 eshayne Exp eshayne $
  $Revision: 1.6 $
 

  $Log: LinkSyncDesktopLastState.cpp,v $
  Revision 1.6  2002-08-10 20:34:32-04  eshayne
  changed name to Link Synchronizer

  Revision 1.5  2002-07-21 10:24:17-04  eshayne
  added comment and license blocks


	----------------------------------------
    Link Synchronizer
    Copyright (C)2002 Ethan Shayne

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "LinkSyncDesktopLastState.h"
#include <fstream.h>
#include "LinkSyncConduitException.h"
#include "LinkSyncDesktopFavorite.h"
#include <SyncMgr.h>
#include <HSLog.h>


LinkSyncDesktopLastState::LinkSyncDesktopLastState(const char* pszDirectory)
: mstrDirectory(pszDirectory)
{
}


LinkSyncDesktopLastState::LinkSyncDesktopLastState(const char* pszDirectory, const LinkSyncDesktopFavorite::FavoriteList& colFavorites)
: mstrDirectory(pszDirectory)
{
	LinkSyncDesktopFavorite::FavoriteList::const_iterator iterFav;
	unsigned long ulYieldCount = 0;
	long err = 0;
	for (iterFav = colFavorites.begin(); iterFav != colFavorites.end(); iterFav++)
	{
		if (++ulYieldCount >= 1000)
		{
			ulYieldCount = 0;
			err = SyncYieldCycles(1);
			if (err != 0)
			{
				LogAddFormattedEntry(slWarning, true, "Could not yeild cycles during while initializing instance of LinkSyncDesktopLastState, err = %d", err);
			}
		}

		mcolFavorites[(*iterFav).GetName()] = (*iterFav);
	}
}


LinkSyncDesktopLastState::~LinkSyncDesktopLastState()
{
}


void LinkSyncDesktopLastState::LoadFromFile()
{
	// open up the file on the desktop which holds the data about the
	// last state of the desktop favorites as of the last sync with this
	// desktop machine. loop through the file and load in info about each
	// favorite, as the desktop knew it at that point in time.

	ifstream fLastState(GetFullFilename().c_str(), ios::in);
	if (!fLastState)
	{
		DWORD dwErr = GetLastError();
		stlport::string strErr = "Could not load last desktop state file from full filename " + GetFullFilename();
		throw LinkSyncConduitException(__FILE__, __LINE__, strErr.c_str(), dwErr);
	}

	char acBuffer[30960];
	int iBufferSize = 30960;
	stlport::string strRecord;
	stlport::string strName;
	stlport::string strUrl;
	stlport::string strDirectory;
	stlport::string strHHRecordId;
	size_t uiDelimiterPos = 0;
	size_t uiStartPos = 0;
	unsigned long ulYieldCount = 0;
	long err = 0;
	while (fLastState.getline(acBuffer, iBufferSize, LINKSYNC_DESKTOP_LASTSTATE_RECORD_DELIMITER))
	{
		if (++ulYieldCount >= 100)
		{
			ulYieldCount = 0;
			err = SyncYieldCycles(1);
			if (err != 0)
			{
				LogAddFormattedEntry(slWarning, true, "Could not yeild cycles during sync while loading LastDesktopState records from file, err = %d", err);
			}
		}

		strRecord = acBuffer;
		// the record format in this file is <name><DELIMITER><url><DELIMITER><desktop directory><DELIMITER><palm record id><RECORD_DELIMITER>
		// extract the individual pieces for this record
		uiDelimiterPos = strRecord.find(LINKSYNC_DESKTOP_LASTSTATE_FIELD_DELIMITER, 0);
		if (uiDelimiterPos == stlport::string::npos)
		{
			stlport::string strErr = "could not parse last desktop state record: '" + strRecord + "'";
			throw LinkSyncConduitException(__FILE__, __LINE__, strErr.c_str());
		}
		strName = strRecord.substr(0, uiDelimiterPos);

		uiStartPos = uiDelimiterPos + 1;
		uiDelimiterPos = strRecord.find(LINKSYNC_DESKTOP_LASTSTATE_FIELD_DELIMITER, uiStartPos);
		if (uiDelimiterPos == stlport::string::npos)
		{
			stlport::string strErr = "could not parse last desktop state record: '" + strRecord + "'";
			throw LinkSyncConduitException(__FILE__, __LINE__, strErr.c_str());
		}
		strUrl = strRecord.substr(uiStartPos, (uiDelimiterPos - uiStartPos));

		uiStartPos = uiDelimiterPos + 1;
		uiDelimiterPos = strRecord.find(LINKSYNC_DESKTOP_LASTSTATE_FIELD_DELIMITER, uiStartPos);
		if (uiDelimiterPos == stlport::string::npos)
		{
			stlport::string strErr = "could not parse last desktop state record: '" + strRecord + "'";
			throw LinkSyncConduitException(__FILE__, __LINE__, strErr.c_str());
		}
		strDirectory = strRecord.substr(uiStartPos, (uiDelimiterPos - uiStartPos));
	
		strHHRecordId = strRecord.substr(uiDelimiterPos + 1);

		// put all the pieces into an instance of LinkSyncDesktopFavorite
		LinkSyncDesktopFavorite fav;
		fav.SetName(strName.c_str());
		fav.SetUrl(strUrl.c_str());
		fav.SetDirectory(strDirectory.c_str());
		DWORD dwHHRecordId = strtoul(strHHRecordId.c_str(), NULL, 10);
		fav.SetHHRecordId(dwHHRecordId);

		// add this favorite to the list
		mcolFavorites[strName] = fav;
	}
}


void LinkSyncDesktopLastState::WriteToFile()
{
	// replace the existing file
	ofstream fLastState(GetFullFilename().c_str(), ios::out | ios::trunc);
	if (!fLastState)
	{
		DWORD dwErr = GetLastError();
		stlport::string strErr = "Could not open last desktop state file for writing: full filename " + GetFullFilename();
		throw LinkSyncConduitException(__FILE__, __LINE__, strErr.c_str(), dwErr);
	}

	// loop through the favorites in memory and write out each one
	// the record format in this file is <name><DELIMITER><url><DELIMITER><desktop directory><DELIMITER><palm record id><RECORD_DELIMITER>
	LinkSyncDesktopFavorite::FavoriteMap::const_iterator iterFav;
	stlport::string strRecord;
	char acHHRecordId[1024];
	unsigned long ulYieldCount = 0;
	long err = 0;
	stlport::string strDelimiter(1, LINKSYNC_DESKTOP_LASTSTATE_FIELD_DELIMITER);
	stlport::string strRecordDelimiter(1, LINKSYNC_DESKTOP_LASTSTATE_RECORD_DELIMITER);
	for (iterFav = mcolFavorites.begin(); iterFav != mcolFavorites.end(); iterFav++)
	{
		if (++ulYieldCount >= 100)
		{
			ulYieldCount = 0;
			err = SyncYieldCycles(1);
			if (err != 0)
			{
				LogAddFormattedEntry(slWarning, true, "Could not yeild cycles during sync while writing LastDesktopState records to file, err = %d", err);
			}
		}

		const LinkSyncDesktopFavorite& fav = (*iterFav).second;
		ultoa(fav.GetHHRecordId(), acHHRecordId, 10);
		strRecord = stlport::string(fav.GetName()) + strDelimiter + stlport::string(fav.GetUrl()) + strDelimiter + stlport::string(fav.GetDirectory()) + strDelimiter + stlport::string(acHHRecordId);
		fLastState << strRecord << strRecordDelimiter << flush;
	}
}


stlport::string LinkSyncDesktopLastState::GetFullFilename() const
{
	return (mstrDirectory + "\\LinkSyncLastDesktopState.txt");
}


LinkSyncDesktopFavorite::FavoriteMap& LinkSyncDesktopLastState::GetFavorites()
{
	return mcolFavorites;
}


bool LinkSyncDesktopLastState::IsEmpty() const
{
	return mcolFavorites.empty();
}


