/******************************************************************************

  $Header: f:\\RCS\\F\\devel\\pilot\\cppconduit\\LinkSyncConduit\\linksyncconduitgencond.cpp,v 1.17 2002-08-30 21:45:27-04 eshayne Exp eshayne $
  $Revision: 1.17 $
 

  $Log: linksyncconduitgencond.cpp,v $
  Revision 1.17  2002-08-30 21:45:27-04  eshayne
  - added subfolder support
  - fixed bug in LinkSyncDesktopFavorite::GetAllFavorites(): was leaving file handles open

  Revision 1.16  2002-08-10 20:34:33-04  eshayne
  changed name to Link Synchronizer

  Revision 1.15  2002-07-21 10:24:18-04  eshayne
  added comment and license blocks


	----------------------------------------
    Link Synchronizer
    Copyright (C)2002 Ethan Shayne

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 *
 * EXPORTED FUNCTIONS Generic Conduit Entry Points Source File
 *
 ****************************************************************************/
#define STRICT 1
#define ASSERT(f)          ((void)0)
#define TRACE0(sz)
#define TRACE(sz)
#include <windows.h>
#include <string.h>
#include <stdio.h>
#ifdef METROWERKS_WIN
#include <wmem.h>
#else
#include <memory.h>
#endif
#include <sys/stat.h>
#include <TCHAR.H>
#include <COMMCTRL.H>


#include <syncmgr.h>
#include "LinkSyncConduitGenCond.h"
#include "LinkSyncConduitException.h"
#include "LinkSyncDesktopFavorite.h"
#include "LinkSyncDesktopLastState.h"

#include <logstrng.h>
#include <HSLog.h>


#include <PALM_CMN.H>
#include "resource.h"

HANDLE hLangInstance;
HANDLE hAppInstance;
extern HANDLE hLangInstance;
extern HANDLE hAppInstance;

long CALLBACK ConfigureDlgProc(HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam);
void LoadCfgDlgBitmaps(HWND hDlg);

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////
//
//     Function:        DllMain()
//
//     Description:    main entry point to the LinkSyncConduit component
//
//     Parameters:    hInstance - instance handle of the DLL
//                    dwReason  - why the entry point was called
//                    lpReserved - reserved
//
//     Returns:        1 if okay
//
/////////////////////////////////////////////////////////////////////////////
//     REVISION HISTORY:
//      03/25/98 KRM inital 
//
/////////////////////////////////////////////////////////////////////////////
#ifndef _68K_

static int iTerminationCount = 0;

extern "C" int APIENTRY
DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{

    if (dwReason == DLL_PROCESS_ATTACH)
    {
        TRACE0("EXPORTED FUNCTIONS Initializing!\n");
        
        if (!iTerminationCount ) {
            hAppInstance = hInstance;
			// use PalmLoadLanguage here to load different languages
            hLangInstance = hInstance;
		        }
        ++iTerminationCount;
    }
    else if (dwReason == DLL_PROCESS_DETACH)
    {
        TRACE0("EXPORTED FUNCTIONS Terminating!\n");

        --iTerminationCount;
        if (!iTerminationCount ) {
			// use PalmFreeLanguage here to unload different languages
        }
    }
    return 1;   // ok
}
#endif


/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////
//
//     Function:        OpenConduit()
//
//     Description:  Extern "C" entry point into this conduit which starts the
//                 process of synchronizing the local database table with
//                 a remote conterpart residing on the remote view device. 
//
//     Parameters:   Pointer to a callback function used to report progress.
//                    
//                
//
//     Returns:        
//
/////////////////////////////////////////////////////////////////////////////
//     REVISION HISTORY:
//      03/25/98 KRM inital 
//
/////////////////////////////////////////////////////////////////////////////
ExportFunc long OpenConduit(PROGRESSFN pFn, CSyncProperties& rProps)
{
    long retval = -1;
    if (pFn)
    {
		// what's the name of this conduit?
		char conduitName[100];
		GetConduitName(conduitName, sizeof(conduitName));

		// if this conduit is configured to do nothing, then do nothing
		// (but log that we're doing nothing and why)
		if (rProps.m_SyncType == eDoNothing)
		{
			LogAddFormattedEntry(slText, true, "%s - sync configured to Do Nothing", conduitName);
			return 0;
		}

		// otherwise, first thing is register this conduit and get a handle
		// that we need to use later to access the conduit manager
		CONDHANDLE hConduit = (CONDHANDLE)0;
		long retval = SyncRegisterConduit(hConduit);
		if (retval != 0)
		{
			// conduit registration failed
			LogAddFormattedEntry(slWarning, true, "%s - sync failed - conduit could not be registered", conduitName);
			return retval;
		}

		// which kind of sync are we doing?
		switch (rProps.m_SyncType)
		{
			case ePCtoHH:
				retval = SyncPCtoHH(rProps);
				break;

			case eHHtoPC:
				retval = SyncHHtoPC(rProps);
				break;

			case eFast:
			case eSlow:
				retval = SyncMirror(rProps);
				break;

			default:
				LogAddFormattedEntry(slWarning, true, "%s - sync failed - conduit does not support selected sync mechanism", conduitName);
				retval = CONDERR_BAD_SYNC_TYPE;
				break;				
		}

		// leave a note in the log indicating whether this sync was successful or not
		LogAddEntry(conduitName, retval ? slSyncAborted : slSyncFinished, false);

		// cleanup and exit
		SyncUnRegisterConduit(hConduit);
		return retval;

    }
    return(retval);
}


// do a full sync, combining any changes on the desktop and the handheld
long SyncMirror(CSyncProperties& rProps)
{
	// because the desktop records are out of our control, we have no way
	// of knowing if any particular "record" (file) on the desktop has
	// changed, or been marked as deleted, etc. We also have no reliable
	// way of even identifying which records on the handheld reflect which
	// records on the desktop. So each sync we need to compare each and
	// every record on the handheld with each and every record on the desktop
	// and then make whatever adjustments are necessary

	long err = 0;
	bool dbOpen = false;
	BYTE rHandle = NULL;
	try {
		// first, load in all the records from the desktop into a map (for
		// easy lookup)
		LinkSyncDesktopFavorite::FavoriteList colPCFavoriteList;
		LinkSyncDesktopFavorite::GetAllFavorites(colPCFavoriteList);
		LinkSyncDesktopFavorite::FavoriteMap colPCFavoriteMap;
		LinkSyncDesktopFavorite::PutFavoritesIntoMap(colPCFavoriteList, colPCFavoriteMap);

		// next, load all the records from the desktop "last state" file to
		// get a list of what the desktop state was the previous time we synced
		// with this desktop machine
		LinkSyncDesktopLastState lastDesktopState(rProps.m_PathName);
		lastDesktopState.LoadFromFile();
		LinkSyncDesktopFavorite::FavoriteMap& colPCLastStateFavoriteMap = lastDesktopState.GetFavorites();

		// next, load in all the records from the handheld into the same kind
		// of structure
		rHandle = OpenDB(false);
		LinkSyncDesktopFavorite::FavoriteList colHHFavoriteList;
		LoadAllHandheldRecords(colHHFavoriteList, rHandle);
		LinkSyncDesktopFavorite::FavoriteMap colHHFavoriteMap;
		LinkSyncDesktopFavorite::PutFavoritesIntoMap(colHHFavoriteList, colHHFavoriteMap);

		// compare the records in both structures and get out these lists:
		// 1. records needing to be added to the handheld (were newly added on the desktop)
		// 2. records needing to be added to the desktop (were newly added on the handheld)
		// 3. records needing to be deleted from the handheld (were deleted on the desktop)
		// 4. records needing to be deleted from the desktop (were deleted on the handheld)
		// 5. records needing to be modified on the handheld (record identified as the matching one was changed on the desktop)
		//			note that this applies to url changes only. anything where the name changed,
		//			the old record should be marked for deletion and the new record marked for adding.
		// 6. records needing to be modified on the desktop (record identified as the matching one was changed on the handheld)
		LinkSyncDesktopFavorite::FavoriteList colAddToHH;
		LinkSyncDesktopFavorite::FavoriteList colAddToPC;
		LinkSyncDesktopFavorite::FavoriteList colDeleteFromHH;
		LinkSyncDesktopFavorite::FavoriteList colDeleteFromPC;
		LinkSyncDesktopFavorite::FavoriteList colChangeOnHH;
		LinkSyncDesktopFavorite::FavoriteList colChangeOnPC;
		LinkSyncDesktopFavorite::FavoriteList colNewPCLastState;
		CompareRecords(colPCFavoriteMap,
					   colPCLastStateFavoriteMap,
					   colHHFavoriteMap,
					   colAddToHH,
					   colAddToPC,
					   colDeleteFromHH,
					   colDeleteFromPC,
					   colChangeOnHH,
					   colChangeOnPC,
					   colNewPCLastState);

		// we've got our marching orders, go make it all happen
		// make any handheld changes, then close the handheld database
		AddRecordsToHandheld(colAddToHH, rHandle);
		DeleteRecordsFromHandheld(colDeleteFromHH, rHandle);
		ChangeRecordsOnHandheld(colChangeOnHH, rHandle);
		// purge all deleted/archived records on the handheld
		err = SyncPurgeDeletedRecs(rHandle);
		if (err != 0)
			throw LinkSyncConduitException(__FILE__, __LINE__, "Could not purge deleted records from the handheld", err);
		CloseDB(rHandle, true);
		dbOpen = false;

		// make any desktop changes
		ChangeRecordsOnDesktop(colChangeOnPC);
		AddRecordsToDesktop(colAddToPC);
		DeleteRecordsFromDesktop(colDeleteFromPC, true);

		// last, we have a collection of records that are intended to represent
		// the new "last desktop state" file. however, this collection has
		// a few records in it with no handheld id, because at the time it was
		// created we didn't know the handheld id. these are records that were
		// added to the handheld via AddRecordsToHandheld() - or more specifically,
		// the records that are also in colAddToHH. so, we need to run through
		// that collection and update the handheld record id's in colNewPCLastState.
		// then we can save it back to the file.
		LinkSyncDesktopLastState newLastState(rProps.m_PathName, colNewPCLastState);
		LinkSyncDesktopFavorite::FavoriteMap& colNewPCLastStateMap = newLastState.GetFavorites();
		LinkSyncDesktopFavorite::FavoriteList::const_iterator iterAddToHH;
		unsigned long ulYieldCount = 0;
		for (iterAddToHH = colAddToHH.begin(); iterAddToHH != colAddToHH.end(); iterAddToHH++)
		{
			const LinkSyncDesktopFavorite& hhAddedFavorite = (*iterAddToHH);
			LinkSyncDesktopFavorite& newLastStateFav = colNewPCLastStateMap[(*iterAddToHH).GetName()];
			newLastStateFav.SetHHRecordId(hhAddedFavorite.GetHHRecordId());
			if (++ulYieldCount >= 100)
			{
				ulYieldCount = 0;
				long lErr = SyncYieldCycles(1);
				if (lErr != 0)
				{
					LogAddFormattedEntry(slWarning, true, "Could not yield cycles during sync, while setting handheld ids on records after mirror sync: error = %d", lErr);
				}
			}
		}
		newLastState.WriteToFile();
	}
	catch (const LinkSyncConduitException& e)
	{
		if (dbOpen)
			CloseDB(rHandle, false);
		LogAddEntry(e.GetErrorText().c_str(), slWarning, true);
		err = e.GetReturnCode();
	}

	return err;
}


void AddRecordsToHandheld(LinkSyncDesktopFavorite::FavoriteList& colFavorites)
{
	BYTE rHandle = NULL;
	bool dbOpen = false;

	// open the database on the handheld
	try {
		rHandle = OpenDB(true);
		dbOpen = true;

		AddRecordsToHandheld(colFavorites, rHandle);
	}
	catch (...)
	{
		if (dbOpen)
			CloseDB(rHandle, false);
		throw;
	}

	// close the handheld database
	if (dbOpen)
		CloseDB(rHandle, true);
}

	
void AddRecordsToHandheld(LinkSyncDesktopFavorite::FavoriteList& colFavorites, BYTE rHandle)
{
	LinkSyncDesktopFavorite::FavoriteList::iterator iterFav;
	long err = 0;
	for (iterFav = colFavorites.begin(); iterFav != colFavorites.end(); iterFav++)
	{
		// convert to a single large character buffer (which is what we'll
		// be storing on the handheld)
		LinkSyncConduitRawData rawData = (*iterFav).GetPalmDBFormattedData();

		// add it to the handheld
		CRawRecordInfo recordInfo;
		recordInfo.m_FileHandle = rHandle;
		// specify an id of 0 to indicate to just append it to the end of the database
		recordInfo.m_RecId = 0;
		recordInfo.m_pBytes = (unsigned char*)rawData.GetData();
		recordInfo.m_Attribs = 0;
		recordInfo.m_CatId = 0;
		recordInfo.m_RecSize = rawData.GetSize();
		recordInfo.m_dwReserved = 0;
		if ((err = SyncWriteRec(recordInfo)) != 0)
		{
			LogAddFormattedEntry(slWarning, true, "Could not add entry to handheld with name %s, url %s, m_RecSize = %d: error = %d", (*iterFav).GetName(), (*iterFav).GetUrl(), recordInfo.m_RecSize, err);
			continue;
		}
		(*iterFav).SetHHRecordId(recordInfo.m_RecId);
	}
}


void AddRecordsToDesktop(LinkSyncDesktopFavorite::FavoriteList& colFavorites)
{
	LinkSyncDesktopFavorite::FavoriteList::iterator iterFav;
	unsigned long ulYieldCount = 0;
	long err = 0;
	for (iterFav = colFavorites.begin(); iterFav != colFavorites.end(); iterFav++)
	{
		(*iterFav).SetDirectory(LinkSyncDesktopFavorite::GetUserFavoritesRootDirectory());
		(*iterFav).WriteToFile();

		if (++ulYieldCount >= 10)
		{
			ulYieldCount = 0;
			err = SyncYieldCycles(1);
			if (err != 0)
			{
				LogAddFormattedEntry(slWarning, true, "Could not yeild cycles during sync while adding records to desktop, err = %d", err);
			}
		}
	}
}


void LoadAllHandheldRecords(LinkSyncDesktopFavorite::FavoriteList& colFavorites)
{
	bool bDBOpen = false;
	BYTE rHandle = OpenDB(false);
	try
	{
		LoadAllHandheldRecords(colFavorites, rHandle);
	}
	catch (...)
	{
		if (bDBOpen)
			CloseDB(rHandle, false);
		throw;
	}

	if (bDBOpen)
		CloseDB(rHandle, true);
}


void LoadAllHandheldRecords(LinkSyncDesktopFavorite::FavoriteList& colFavorites, BYTE rHandle)
{
	colFavorites.erase(colFavorites.begin(), colFavorites.end());
	long err = 0;

	int i = 0;
	CRawRecordInfo recordInfo;
	unsigned short usNameLen = 0;
	char* pszUrlPosInBuffer = NULL;

	// check how many records on the handheld we need to deal with
	WORD recordCount;
	if ((err = SyncGetDBRecordCount(rHandle, recordCount)) != 0)
		throw LinkSyncConduitException(__FILE__, __LINE__, "SyncGetDBRecordCount failed", err);

	// loop through the records on the handheld and do whatever we need
	// to do on the PC for each one
	recordInfo.m_FileHandle = rHandle;
	for (i = 0; i < recordCount; i++)
	{
		LinkSyncConduitRawData rawData;
		rawData.Allocate(1024);
		recordInfo.m_RecIndex = i;
		recordInfo.m_pBytes = (unsigned char*)rawData.GetData();
		recordInfo.m_TotalBytes = rawData.GetAllocatedSize();
		recordInfo.m_dwReserved = 0;

		if ((err = SyncReadRecordByIndex(recordInfo)) != 0)
		{
			LogAddFormattedEntry(slWarning, true, "SyncReadRecordByIndex failed on index # %d with error code = %d", i, err);
			continue;
		}
		if (recordInfo.m_RecSize > recordInfo.m_TotalBytes)
		{
			// m_TotalBytes can only be an unsigned short, even though m_RecSize
			// can be an unsigned long (!), so unfortunately if m_RecSize is
			// larger than what an unsigned short can hold, we're SOL
			if (recordInfo.m_RecSize > USHRT_MAX)
			{
				LogAddFormattedEntry(slWarning, true, "record with index %d is too large to sync (%d)", i, recordInfo.m_RecSize);
				continue;
			}
			rawData.Allocate((unsigned short)recordInfo.m_RecSize);
			recordInfo.m_pBytes = (unsigned char*)rawData.GetData();
			recordInfo.m_TotalBytes = rawData.GetAllocatedSize();
			if ((err = SyncReadRecordByIndex(recordInfo)) != 0)
			{
				LogAddFormattedEntry(slWarning, true, "SyncReadRecordByIndex failed on index # %d with error code = %d; record size = %d, allocated size = %d", i, err, recordInfo.m_RecSize, recordInfo.m_TotalBytes);
				continue;
			}
		}

		// create an instance of LinkSyncDesktopFavorite to represent this
		// record on the desktop
		LinkSyncDesktopFavorite desktopRecord(rawData);
		desktopRecord.SetHHRecordId(recordInfo.m_RecId);
		colFavorites.push_back(desktopRecord);
	}
}


// overwrite whatever's on the desktop with whatever's on the handheld
long SyncHHtoPC(CSyncProperties& rProps)
{
	long err = 0;
	BYTE rHandle = NULL;
	bool dbOpen = false;

	try {
		// get all the handheld records
		LinkSyncDesktopFavorite::FavoriteList colHHFavorites;
		rHandle = OpenDB(true);
		dbOpen = true;
		LoadAllHandheldRecords(colHHFavorites, rHandle);
		// purge all deleted/archived records on the handheld
		err = SyncPurgeDeletedRecs(rHandle);
		if (err != 0)
			throw LinkSyncConduitException(__FILE__, __LINE__, "Could not purge deleted records from the handheld", err);
		CloseDB(rHandle, true);
		dbOpen = false;
	
		// delete all files on the desktop (in the favorites directory that is)
		LinkSyncDesktopFavorite::FavoriteList colPCFavorites;
		LinkSyncDesktopFavorite::GetAllFavorites(colPCFavorites);
		DeleteRecordsFromDesktop(colPCFavorites, false);
		
		// iterate over the records from the handheld and write each one to the desktop
		AddRecordsToDesktop(colHHFavorites);

		// update the desktop last state file
		UpdateDesktopLastStateFile(rProps, colHHFavorites);
	}
	catch (const LinkSyncConduitException& e)
	{
		if (dbOpen)
		{
			CloseDB(rHandle, false);
			dbOpen = false;
		}
		LogAddEntry(e.GetErrorText().c_str(), slWarning, true);
		err = e.GetReturnCode();
	}
	catch (...)
	{
		if (dbOpen)
		{
			CloseDB(rHandle, false);
			dbOpen = false;
		}
		LogAddEntry("unknown exception caught", slWarning, true);
		if (err == 0)
			err = -1;
	}

	return err;
}


// overwrite whatever's on the handheld with whatever's on the desktop
long SyncPCtoHH(CSyncProperties& rProps)
{
	bool dbOpen = false;
	long err = 0;
	BYTE rHandle = NULL;

	// open the database on the handheld
	try {
		rHandle = OpenDB(true);
		dbOpen = true;

		// first thing is to delete all records on the handheld, since this
		// is a one-way "pc overwrites handheld", NOT a synchronization
		if ((err = SyncPurgeAllRecs(rHandle)) != 0)
			throw LinkSyncConduitException(__FILE__, __LINE__, "SyncPurgeAllRecs failed", err);

		// load up all the records from the desktop
		LinkSyncDesktopFavorite::FavoriteList colFavorites;
		LinkSyncDesktopFavorite::GetAllFavorites(colFavorites);

		// iterate over the desktop records and for each one, add a handheld record
		AddRecordsToHandheld(colFavorites, rHandle);

		// update the "last desktop state" file
		UpdateDesktopLastStateFile(rProps, colFavorites);
	}
	catch (const LinkSyncConduitException& e)
	{
		LogAddEntry(e.GetErrorText().c_str(), slWarning, true);
		err = e.GetReturnCode();
	}

	// close the handheld database
	if (dbOpen)
		CloseDB(rHandle, false);

	return err;
}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////
//
//       Function:              GetConduitName()
//
//       Description:  Extern "C" entry point into this conduit which returns
//                                 the name to be used when display messages regarding
//                                 this conduit.
//
//       Parameters:   pszName - buffer in which to place the name
//                                 nLen - maximum number of bytes of buffer     
//                                      
//                              
//
//       Returns:          -1 indicates erros
//
/////////////////////////////////////////////////////////////////////////////
//       REVISION HISTORY:
//      03/25/98 KRM inital 
//
/////////////////////////////////////////////////////////////////////////////
ExportFunc long GetConduitName(char* pszName,WORD nLen)
{
    long retval = -1;

    if (::LoadString((HINSTANCE)hLangInstance, IDS_CONDUIT_NAME, pszName, nLen))
        retval = 0;

    return retval;
}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////
//
//       Function:     GetConduitVersion()
//
//       Description:  Extern "C" entry point into this conduit which returns
//                     the conduits version
//
//       Parameters:   none
//
//       Returns:      DWORD indicating major and minor version number
//                        HIWORD - reserved
//                        HIBYTE(LOWORD) - major number
//                        LOBYTE(LOWORD) - minor number
//
/////////////////////////////////////////////////////////////////////////////
//       REVISION HISTORY:
//      03/25/98 KRM inital 
/////////////////////////////////////////////////////////////////////////////
ExportFunc DWORD GetConduitVersion()
{
    return GENERIC_CONDUIT_VERSION;
}

/////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////
//
//       Function:     ConfigureConduit
//
//       Description:  Extern "C" entry point into this conduit which returns
//                     should display the UI necessary to configure this 
//                       conduit.
//
//       Parameters:   none
//
//       Returns:      0 - success, !0 - failure
//
/////////////////////////////////////////////////////////////////////////////
//       REVISION HISTORY:
//      03/25/98 KRM inital 
/////////////////////////////////////////////////////////////////////////////
ExportFunc long ConfigureConduit(CSyncPreference& pref)
{

    long nRtn = -1;
    CfgConduitInfoType cfg;
    cfg.dwVersion = CFGCONDUITINFO_VERSION_1;
    cfg.dwSize  = sizeof(CfgConduitInfoType);
    cfg.dwCreatorId = 0;
    cfg.dwUserId = 0;
    memset(cfg.szUser , 0, sizeof(cfg.szUser));
    memset(cfg.m_PathName, 0, sizeof(cfg.m_PathName)); 
    cfg.syncPermanent = pref.m_SyncType;
    cfg.syncTemporary = pref.m_SyncType;
    cfg.syncNew = pref.m_SyncType;
    cfg.syncPref = eTemporaryPreference; 

    int iResult;
    iResult = DialogBoxParam((HINSTANCE)hLangInstance, 
              MAKEINTRESOURCE(IDD_CONDUIT_ACTION), 
              GetForegroundWindow(), 
              (DLGPROC)ConfigureDlgProc,
              (LPARAM)&cfg);
    if (iResult == 0) {
        pref.m_SyncType = cfg.syncNew;
        pref.m_SyncPref = cfg.syncPref;
        nRtn = 0;
    }
    return nRtn;
}

/////////////////////////////////////////////////////////////////////////////
// 
/////////////////////////////////////////////////////////////////////////////
//
//	 Method:		GetConduitInfo
//
//	 Description:	This function provides a way for a Conduit to provide info
//                  to the caller. 
//                  In this version of the call, MFC Version, Conduit Name, and
//                  Default sync action are the types of information this call
//                  will return.
//
//	 Parameters:	ConduitInfoEnum infoType - enum specifying what info is being
//                          requested.
//                  void *pInArgs - This parameter may be null, except for the Conduit
//                          name enum, this value will be a ConduitRequestInfoType structure.
//                  This following to parameters vary depending upon the info being requested. 
//                  For enum eConduitName
//                  void *pOut - will be a pointer to a character buffer
//                  DWORD *pdwOutSize - will be a pointer to a DWORD specifying the size of the character buffer.
//
//                  For enum eMfcVersion
//                  void *pOut - will be a pointer to a DWORD
//                  DWORD *pdwOutSize - will be a pointer to a DWORD specifying the size of pOut.
//
//                  For enum eDefaultAction
//                  void *pOut - will be a pointer to a eSyncType variable
//                  DWORD *pdwOutSize - will be a pointer to a DWORD specifying the size of pOut.
//
//	 Returns:		0		- Success.
//					!0      - error code.
//
/////////////////////////////////////////////////////////////////////////////
//	 REVISION HISTORY:
//		01/23/98    KRM Created
//
/////////////////////////////////////////////////////////////////////////////
ExportFunc long GetConduitInfo(ConduitInfoEnum infoType, void *pInArgs, void *pOut, DWORD *pdwOutSize)
{
    if (!pOut)
        return CONDERR_INVALID_PTR;
    if (!pdwOutSize)
        return CONDERR_INVALID_OUTSIZE_PTR;

    switch (infoType) {
        case eConduitName:

            // This code is for example. This conduit does not use this code
            
            if (!pInArgs)
                return CONDERR_INVALID_INARGS_PTR;
            ConduitRequestInfoType *pInfo;
            pInfo = (ConduitRequestInfoType *)pInArgs;
            if ((pInfo->dwVersion != CONDUITREQUESTINFO_VERSION_1) ||
                (pInfo->dwSize != SZ_CONDUITREQUESTINFO))
                return CONDERR_INVALID_INARGS_STRUCT;
            
	        if (!::LoadString((HINSTANCE)hLangInstance, IDS_CONDUIT_NAME, (TCHAR*)pOut, *pdwOutSize))
                return CONDERR_CONDUIT_RESOURCE_FAILURE;
            break;
        case eDefaultAction:
            if (*pdwOutSize != sizeof(eSyncTypes))
                return CONDERR_INVALID_BUFFER_SIZE;
            (*(eSyncTypes*)pOut) = eFast;
            break;
        case eMfcVersion:
            if (*pdwOutSize != sizeof(DWORD))
                return CONDERR_INVALID_BUFFER_SIZE;
            (*(DWORD*)pOut) = MFC_NOT_USED;
            break;
        default:
            return CONDERR_UNSUPPORTED_CONDUITINFO_ENUM;
    }
    return 0;
}


////////////////////////////////////===============================///////////

/////////////////////////////////////////////////////////////////////////////
// 
/////////////////////////////////////////////////////////////////////////////
//
//	 Method:		CfgConduit
//
//	 Description:	This is the second instance of ConfigureConduit. This function
//                  is provided with more info from the caller.
//                  This version of the function only supports the eConfig1 enum.
//
//	 Parameters:	ConduitCfgEnum cfgType - enum specifying the version of the 
//                      CfgConduit Call.
//                  void *pArgs - In the eConfig1 call, this is a pointer to aCfgConduitInfoType 
//                      structure. This structure provides extra info to the conduit.
//                  DWORD *pdwArgsSize - This is a pointer to a DWORD specifying the size of pArgs.
//
//	 Returns:		0		- Success.
//					!0      - error code.
//
/////////////////////////////////////////////////////////////////////////////
//	 REVISION HISTORY:
//		01/23/98    KRM Created
//
/////////////////////////////////////////////////////////////////////////////

ExportFunc long CfgConduit( ConduitCfgEnum cfgType, void *pArgs, DWORD *pdwArgsSize)
{
	long nRtn = -1;
	TCHAR szName[256];
    DWORD dwNamesize;
    ConduitRequestInfoType infoStruct;
    CfgConduitInfoType *pCfgInfo;


    dwNamesize = sizeof(szName);


    if (!pArgs)
        return CONDERR_INVALID_INARGS_PTR;
    if (!pdwArgsSize)
        return CONDERR_INVALID_ARGSSIZE_PTR;
    if (*pdwArgsSize != SZ_CFGCONDUITINFO) 
        return CONDERR_INVALID_ARGSSIZE;

    if (cfgType != eConfig1)
            return CONDERR_UNSUPPORTED_CFGCONDUIT_ENUM;

    pCfgInfo = (CfgConduitInfoType *)pArgs;
    if (pCfgInfo->dwVersion != CFGCONDUITINFO_VERSION_1)
        return CONDERR_UNSUPPORTED_STRUCT_VERSION;



    infoStruct.dwVersion = CONDUITREQUESTINFO_VERSION_1;
    infoStruct.dwSize = SZ_CONDUITREQUESTINFO;
    infoStruct.dwCreatorId = pCfgInfo->dwCreatorId;
    infoStruct.dwUserId = pCfgInfo->dwUserId;
    strcpy(infoStruct.szUser, pCfgInfo->szUser);
    nRtn = GetConduitInfo(eConduitName, (void *)&infoStruct, (void *)szName, &dwNamesize);
    if (nRtn)
        return nRtn;

    int iResult;
    iResult = DialogBoxParam((HINSTANCE)hLangInstance, 
              MAKEINTRESOURCE(IDD_CONDUIT_CFG_DETAILED), 
              GetForegroundWindow(), 
              (DLGPROC)ConfigureDlgProc,
              (LPARAM)pCfgInfo);

    return 0;
}









long CALLBACK ConfigureDlgProc(HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
    static CfgConduitInfoType *pCfgInfo;
    TCHAR szPath[256];

    switch (Message) {
        case WM_INITDIALOG:

            if (lParam != 0) {
                TCHAR szBuffer[256];
                TCHAR szBuf2[256];
                LoadString((HINSTANCE)hLangInstance, IDS_SYNC_ACTION_TEXT, szBuffer, sizeof(szBuffer));
                SetDlgItemText(hWnd, IDC_ACTIONGROUPBOXTEXT, szBuffer);

                LoadString((HINSTANCE)hLangInstance, IDS_CONDUIT_NAME, szBuffer, sizeof(szBuffer));
                SetWindowText(hWnd, szBuffer);
     
				// Load the bitmaps properly
				LoadCfgDlgBitmaps(hWnd);

                pCfgInfo = (CfgConduitInfoType *)lParam;
                switch (pCfgInfo->syncTemporary){
                    case eFast:
                    case eSlow:
                        CheckRadioButton( hWnd, IDC_RADIO_SYNC, IDC_RADIO_DONOTHING, IDC_RADIO_SYNC);
                        LoadString((HINSTANCE)hLangInstance, IDS_SYNC_FILES, szBuffer, sizeof(szBuffer));
                        break;
                    case ePCtoHH:
                        CheckRadioButton( hWnd, IDC_RADIO_SYNC, IDC_RADIO_DONOTHING, IDC_RADIO_PCTOHH);
                        LoadString((HINSTANCE)hLangInstance, IDS_PCTOHH, szBuffer, sizeof(szBuffer));
                        break;
                    case eHHtoPC:
                        CheckRadioButton( hWnd, IDC_RADIO_SYNC, IDC_RADIO_DONOTHING, IDC_RADIO_HHTOPC);
                        LoadString((HINSTANCE)hLangInstance, IDS_HHTOPC, szBuffer, sizeof(szBuffer));
                        break;
                    case eDoNothing:
                    default:
                        CheckRadioButton( hWnd, IDC_RADIO_SYNC, IDC_RADIO_DONOTHING, IDC_RADIO_DONOTHING);
                        LoadString((HINSTANCE)hLangInstance, IDS_DO_NOTHING, szBuffer, sizeof(szBuffer));
                        break;
                }
                // did we get called from the old config call or the new cfg call?
                if (pCfgInfo->dwCreatorId != 0) {

                    SetDlgItemText(hWnd, IDC_STATIC_TEMPORARY, szBuffer);

                    switch (pCfgInfo->syncPermanent){
                        case eFast:
                        case eSlow:
                            LoadString((HINSTANCE)hLangInstance, IDS_SYNC_FILES, szBuffer, sizeof(szBuffer));
                            break;
                        case ePCtoHH:
                            LoadString((HINSTANCE)hLangInstance, IDS_PCTOHH, szBuffer, sizeof(szBuffer));
                            break;
                        case eHHtoPC:
                            LoadString((HINSTANCE)hLangInstance, IDS_HHTOPC, szBuffer, sizeof(szBuffer));
                            break;
                        case eDoNothing:
                        default:
                            LoadString((HINSTANCE)hLangInstance, IDS_DO_NOTHING, szBuffer, sizeof(szBuffer));
                            break;
                    }
                    SetDlgItemText(hWnd, IDC_STATIC_PERMANENT, szBuffer);

                    LoadString((HINSTANCE)hLangInstance, IDS_CURRENT_SETTINGS_GROUP, szBuffer, sizeof(szBuffer));
                    wsprintf(szBuf2, szBuffer, pCfgInfo->szUser);
                    SetDlgItemText(hWnd, IDC_CURRENT_SETTINGS_GROUP, szBuf2);
                }
            }
            break;
        case WM_COMMAND:
            switch (wParam) {
                case IDC_RADIO_SYNC:
                case IDC_RADIO_PCTOHH:
                case IDC_RADIO_HHTOPC:
                case IDC_RADIO_DONOTHING:
                    CheckRadioButton( hWnd, IDC_RADIO_SYNC, IDC_RADIO_DONOTHING, wParam);
                    break;
                case IDCANCEL:
                    EndDialog(hWnd, 1);
                    return TRUE;
                case IDOK:
                    if (IsDlgButtonChecked(hWnd, IDC_RADIO_SYNC)) {
                        pCfgInfo->syncNew = eFast;
                    } else if (IsDlgButtonChecked(hWnd, IDC_RADIO_PCTOHH)) {
                        pCfgInfo->syncNew = ePCtoHH;
                    } else if (IsDlgButtonChecked(hWnd, IDC_RADIO_HHTOPC)) {
                        pCfgInfo->syncNew = eHHtoPC;
                    } else {
                        pCfgInfo->syncNew = eDoNothing;
                    }
                    if (IsDlgButtonChecked(hWnd, IDC_MAKEDEFAULT)){
                        pCfgInfo->syncPref = ePermanentPreference;
                    } else {
                        pCfgInfo->syncPref = eTemporaryPreference;
                    }
                    EndDialog(hWnd, 0);
                    return TRUE;
				case IDHELP:
#define HELP_FILE "hotsync.hlp"
					if (GetModuleFileName((HINSTANCE)hAppInstance, szPath, sizeof(szPath)))
					{
						// Look for the end of the path
						int		fileOffset;
						fileOffset = strlen(szPath);
						for (; fileOffset > 0; fileOffset--)
						{
							if (szPath[fileOffset] == '\\')
								break;
						}

						if (fileOffset > 0)
						{
							szPath[fileOffset + 1] = 0;
							strcat(szPath, HELP_FILE);
							WinHelp(hWnd, szPath, HELP_CONTEXT, IDD_CONDUIT_ACTION + 0x20000 /*base resource*/);
						}
					}
					return TRUE;
					break;
                default:
                    break;
            }
            break;

		case WM_HELP:
			if (GetModuleFileName((HINSTANCE)hAppInstance, szPath, sizeof(szPath)))
			{
				// Look for the end of the path
				int		fileOffset;
				fileOffset = strlen(szPath);
				for (; fileOffset > 0; fileOffset--)
				{
					if (szPath[fileOffset] == '\\')
						break;
				}

				if (fileOffset > 0)
				{
					szPath[fileOffset + 1] = 0;
					strcat(szPath, HELP_FILE);
					WinHelp(hWnd, szPath, HELP_CONTEXT, IDD_CONDUIT_ACTION + 0x20000 /*base resource*/);
				}
			}
			return TRUE;
		case WM_SYSCOLORCHANGE:
			LoadCfgDlgBitmaps(hWnd);
			break;
        default:
            break;
    }
    return FALSE;
}
/////////////////////////////////////////////////////////////////////////////
// 
/////////////////////////////////////////////////////////////////////////////
//
//	 Method:		LoadCfgDlgBitmaps
//
//	 Description:	This routines translates the bitmap gray background to the
//					button face color set in the control panel/window appearances
//
//	 Parameters:	HWND	- hDlg, handle of the dialog
//
//	 Returns:		none
//
/////////////////////////////////////////////////////////////////////////////
//	 REVISION HISTORY:
//		01/15/99    Carl Created
//
/////////////////////////////////////////////////////////////////////////////
void LoadCfgDlgBitmaps(HWND hDlg)
{
	COLORMAP		colorMap;
	HWND			hwndButton;
	HBITMAP			hBitmap, hOldBitmap;

	//
	// setup the bitmaps
	//
	colorMap.to = GetSysColor(COLOR_BTNFACE);
	colorMap.from = RGB(192,192,192);

	// Sync 
	hBitmap = CreateMappedBitmap((HINSTANCE)hLangInstance, IDB_SYNC, 0, &colorMap, 1);
	// associate the bitmap with the button.
	if ((hwndButton = GetDlgItem(hDlg, IDC_SYNC)) != NULL)
	{
		hOldBitmap = (HBITMAP)SendMessage(hwndButton, STM_SETIMAGE,(WPARAM)IMAGE_BITMAP, (LPARAM)(HANDLE)hBitmap);
		if (hOldBitmap != NULL)
			DeleteObject((HGDIOBJ)hOldBitmap);
	}

	// PC to HH 
	hBitmap = CreateMappedBitmap((HINSTANCE)hLangInstance, IDB_PCTOHH, 0, &colorMap, 1);
	// associate the bitmap with the button.
	if ((hwndButton = GetDlgItem(hDlg, IDC_PCTOHH)) != NULL)
	{
		hOldBitmap = (HBITMAP)SendMessage(hwndButton, STM_SETIMAGE,(WPARAM)IMAGE_BITMAP, (LPARAM)(HANDLE)hBitmap);
		if (hOldBitmap != NULL)
			DeleteObject((HGDIOBJ)hOldBitmap);
	}


	// HH to PC 
	hBitmap = CreateMappedBitmap((HINSTANCE)hLangInstance, IDB_HHTOPC, 0, &colorMap, 1);
	// associate the bitmap with the button.
	if ((hwndButton = GetDlgItem(hDlg, IDC_HHTOPC)) != NULL)
	{
		hOldBitmap = (HBITMAP)SendMessage(hwndButton, STM_SETIMAGE,(WPARAM)IMAGE_BITMAP, (LPARAM)(HANDLE)hBitmap);
		if (hOldBitmap != NULL)
			DeleteObject((HGDIOBJ)hOldBitmap);
	}

	// Do Nothing 
	hBitmap = CreateMappedBitmap((HINSTANCE)hLangInstance, IDB_DONOTHING, 0, &colorMap, 1);
	// associate the bitmap with the button.
	if ((hwndButton = GetDlgItem(hDlg, IDC_DONOTHING)) != NULL)
	{
		hOldBitmap = (HBITMAP)SendMessage(hwndButton, STM_SETIMAGE,(WPARAM)IMAGE_BITMAP, (LPARAM)(HANDLE)hBitmap);
		if (hOldBitmap != NULL)
			DeleteObject((HGDIOBJ)hOldBitmap);
	}
}


BYTE OpenDB(const bool bCreateIfNeeded)
{
	BYTE rHandle = NULL;
	long err = 0;
	if ((err = SyncOpenDB("Database-LSnc", 0, rHandle, eDbWrite | eDbRead | eDbShowSecret)) != 0)
	{
		// if the error we got was that the database was not found on the handheld,
		// then create the database on the handheld (maybe just the first time
		// this app has been used)
		if (err == SYNCERR_FILE_NOT_FOUND && bCreateIfNeeded)
		{
			CDbCreateDB dbInfo;
			memset(&dbInfo, 0, sizeof(dbInfo));
			dbInfo.m_Creator = 'LSnc';
			dbInfo.m_Flags = eRecord;
			dbInfo.m_CardNo = 0;
			dbInfo.m_Type = 'DATA';
			strcpy(dbInfo.m_Name, "Database-LSnc");
			if ((err = SyncCreateDB(dbInfo)) != 0)
				throw LinkSyncConduitException(__FILE__, __LINE__, "SyncCreateDB failed", err);
			rHandle = dbInfo.m_FileHandle;
		}
		else
			throw LinkSyncConduitException(__FILE__, __LINE__, "SyncOpenDB failed", err);
	}

	return rHandle;
}


void CloseDB(BYTE rHandle, const bool bThrowException)
{
	long err = 0;
	if ((err = SyncCloseDB(rHandle)) != 0)
	{
		if (bThrowException)
			throw LinkSyncConduitException(__FILE__, __LINE__, "Could not close handheld database", err);
		else
			LogAddEntry("Could not close handheld database", slWarning, true);
	}
}


void DeleteRecordsFromDesktop(LinkSyncDesktopFavorite::FavoriteList& colFavorites,
							  const bool bContinueIfError)
{
	LinkSyncDesktopFavorite::FavoriteList::iterator iterFav;
	DWORD dwErr = 0;
	unsigned long ulYieldCount = 0;
	for (iterFav = colFavorites.begin(); iterFav != colFavorites.end(); iterFav++)
	{
		if (++ulYieldCount >= 10)
		{
			ulYieldCount = 0;
			dwErr = SyncYieldCycles(1);
			if (dwErr != 0)
			{
				LogAddFormattedEntry(slWarning, true, "Could not yeild cycles during sync while deleting records from desktop, err = %d", dwErr);
			}
		}

		try {
			dwErr = (*iterFav).DeleteDesktopFile();
			if (dwErr != 0)
			{
				stlport::string strErr = "Could not delete desktop file " + (*iterFav).GetFilename();
				throw LinkSyncConduitException(__FILE__, __LINE__, strErr.c_str(), dwErr);
			}
		}
		catch (const LinkSyncConduitException& e)
		{
			if (bContinueIfError)
			{
				LogAddEntry(e.GetErrorText().c_str(), slWarning, true);
				continue;
			}
			else
				throw;
		}
	}		
}


void DeleteRecordsFromHandheld(LinkSyncDesktopFavorite::FavoriteList& colFavorites, BYTE rHandle)
{
	LinkSyncDesktopFavorite::FavoriteList::const_iterator iterFav;
	DWORD dwHHRecordId = 0;
	long err = 0;
	CRawRecordInfo rInfo;
	rInfo.m_FileHandle = rHandle;
	rInfo.m_dwReserved = NULL;
	for (iterFav = colFavorites.begin(); iterFav != colFavorites.end(); iterFav++)
	{
		try {
			dwHHRecordId = (*iterFav).GetHHRecordId();
			rInfo.m_RecId = dwHHRecordId;
			err = SyncDeleteRec(rInfo);
			if (err != 0)
				throw LinkSyncConduitException(__FILE__, __LINE__, "Error deleting handheld record", err);
		}
		catch (const LinkSyncConduitException& e)
		{
			stlport::string strErr = "Could not delete handheld record id ";
			char acBuffer[1024];
			ultoa(dwHHRecordId, acBuffer, 10);
			strErr += stlport::string(acBuffer);
			strErr += ": " + e.GetErrorText();
			LogAddEntry(strErr.c_str(), slWarning, true);
			continue;
		}
	}
}


void ChangeRecordsOnHandheld(LinkSyncDesktopFavorite::FavoriteList& colFavorites, BYTE rHandle)
{
	LinkSyncDesktopFavorite::FavoriteList::iterator iterFav;
	DWORD dwHHRecordId = 0;
	CRawRecordInfo rInfo;
	rInfo.m_FileHandle = rHandle;
	rInfo.m_dwReserved = NULL;
	rInfo.m_Attribs = 0;
	rInfo.m_CatId = 0;
	long err = 0;
	for (iterFav = colFavorites.begin(); iterFav != colFavorites.end(); iterFav++)
	{
		try {
			LinkSyncConduitRawData rawData = (*iterFav).GetPalmDBFormattedData();
			dwHHRecordId = (*iterFav).GetHHRecordId();
			// by specifying the record id, we tell SyncWriteRec to overwrite
			// an existing record instead of simply adding a new one. this way
			// not only to we avoid having to do a delete first, but we also
			// preserve ordering in the handheld database
			rInfo.m_RecId = dwHHRecordId;
			rInfo.m_pBytes = (unsigned char*)rawData.GetData();
			rInfo.m_RecSize = rawData.GetSize();
			err = SyncWriteRec(rInfo);
			if (err != 0)
				throw LinkSyncConduitException(__FILE__, __LINE__, "Could not modify record", err);
		}
		catch (const LinkSyncConduitException& e)
		{
			stlport::string strErr = "Could not update record id ";
			char acBuffer[1024];
			ultoa(dwHHRecordId, acBuffer, 10);
			strErr += stlport::string(acBuffer);
			strErr += ": " + e.GetErrorText();
			LogAddEntry(strErr.c_str(), slWarning, true);
			continue;
		}
	}
}


void ChangeRecordsOnDesktop(LinkSyncDesktopFavorite::FavoriteList& colFavorites)
{
	// note that this applies to url changes only. anything where the name changed,
	// the old record was marked for deletion and the new record marked for adding.

	LinkSyncDesktopFavorite::FavoriteList::const_iterator iterFav;
	unsigned long ulYieldCount = 0;
	long err = 0;
	for (iterFav = colFavorites.begin(); iterFav != colFavorites.end(); iterFav++)
	{
		if (++ulYieldCount >= 10)
		{
			ulYieldCount = 0;
			err = SyncYieldCycles(1);
			if (err != 0)
			{
				LogAddFormattedEntry(slWarning, true, "Could not yeild cycles during sync while modifying records on desktop, err = %d", err);
			}
		}

		try {
			// just write out the record - because the name is the same, the filename
			// will also be the same, so it'll just overwrite the old one
			(*iterFav).WriteToFile();
		}
		catch (const LinkSyncConduitException& e)
		{
			stlport::string strErr = "Could not update desktop record: " + e.GetErrorText();
			LogAddEntry(strErr.c_str(), slWarning, true);
			continue;
		}
	}
}


void CompareRecords(const LinkSyncDesktopFavorite::FavoriteMap& colPCFavoriteMap,
					const LinkSyncDesktopFavorite::FavoriteMap& colPCLastStateFavoriteMap,
					const LinkSyncDesktopFavorite::FavoriteMap& colHHFavoriteMap,
					LinkSyncDesktopFavorite::FavoriteList& colAddToHH,
					LinkSyncDesktopFavorite::FavoriteList& colAddToPC,
					LinkSyncDesktopFavorite::FavoriteList& colDeleteFromHH,
					LinkSyncDesktopFavorite::FavoriteList& colDeleteFromPC,
					LinkSyncDesktopFavorite::FavoriteList& colChangeOnHH,
					LinkSyncDesktopFavorite::FavoriteList& colChangeOnPC,
					LinkSyncDesktopFavorite::FavoriteList& colNewPCLastState)
{
	// first, loop through the desktop records and see which ones do or don't 
	// exist at all on the handheld (by "exist" we mean something with the same
	// name)
	LinkSyncDesktopFavorite::FavoriteMap::const_iterator iterPCFavorites;
	LinkSyncDesktopFavorite::FavoriteMap::const_iterator iterHHFavorites;
	LinkSyncDesktopFavorite::FavoriteMap::const_iterator iterPCLastStateFavorites;
	unsigned long ulYieldCount = 0;
	long err = 0;
	for (iterPCFavorites = colPCFavoriteMap.begin(); iterPCFavorites != colPCFavoriteMap.end(); iterPCFavorites++)
	{
		if (++ulYieldCount >= 100)
		{
			ulYieldCount = 0;
			err = SyncYieldCycles(1);
			if (err != 0)
			{
				LogAddFormattedEntry(slWarning, true, "Could not yeild cycles during sync while comparing records from pc to handheld, err = %d", err);
			}
		}

		const stlport::string& strPCName = (*iterPCFavorites).first;
		const LinkSyncDesktopFavorite& pcFavorite = (*iterPCFavorites).second;

		// does it exist on the handheld?
		iterHHFavorites = colHHFavoriteMap.find(strPCName);
		if (iterHHFavorites == colHHFavoriteMap.end())
		{
			// no, it does not exist on the handheld
			// now the question is - was it deleted from the handheld,
			// meaning we need to also delete it on the desktop? or was
			// it newly added on the desktop, meaning we need to add it
			// to the handheld?
			// the answer can be found by looking at the "last desktop state"
			// collection: if this record also exists there (by name, anyway),
			// then we know that it used to exist on the handheld and was
			// deleted there, so therefore it needs to be deleted on the
			// desktop. if this record does *not* exist in the "last desktop
			// state" file, then we know that this entry was added to the
			// desktop (since the last sync) and therefore needs to be added
			// to the handheld
			iterPCLastStateFavorites = colPCLastStateFavoriteMap.find(strPCName);
			if (iterPCLastStateFavorites == colPCLastStateFavoriteMap.end())
			{
				// no, it did not exist on the desktop as of the last sync,
				// so therefore it needs to be added to the handheld
				LinkSyncDesktopFavorite modFavorite = pcFavorite;
				modFavorite.SetHHRecordId(0);
				colAddToHH.push_back(modFavorite);
				colNewPCLastState.push_back(modFavorite);
			}
			else
			{
				// yes, it did exist on the desktop as of the last sync,
				// so therefore it needs to be deleted from the desktop
				colDeleteFromPC.push_back(pcFavorite);
			}
		}
		else
		{
			// yes, it does exist on the handheld. is it the same url?
			const LinkSyncDesktopFavorite& hhFavorite = (*iterHHFavorites).second;
			if (strcmp(hhFavorite.GetUrl(), pcFavorite.GetUrl()) == 0)
			{
				// they are the same - just add it to the new "last state" collection,
				// nothing else to do
				// note - use we need both the pc's directory and the hh's record id
				LinkSyncDesktopFavorite modFavorite = hhFavorite;
				modFavorite.SetDirectory(pcFavorite.GetDirectory());
				colNewPCLastState.push_back(modFavorite);
			}
			else
			{
				// different urls - this record has changed!
				// so now the question is - which one is right?
				// unfortunately, we don't have a lock-tight way of determining
				// this - we just have to take our best guess.
				// if we look at the "last desktop state" file, and it agrees
				// with the desktop, then that means that the handheld
				// changed since the last sync, which means the handheld is most
				// likely the more recent change, so it wins.
				// if the "last desktop state" file agrees with the handheld,
				// then that means that the desktop changed since the last sync,
				// which means the desktop is most likely the more recent change,
				// so it wins.
				// if the "last desktop state" file disagrees with both handheld
				// and desktop, that means both platforms have changed since the
				// last sync, so we take the standard Palm approach and duplicate
				// each version on both platforms and warn the user.
				// if the "last desktop state" file has no record of a record
				// using this name, then it was added on both platforms since
				// the last sync, so again we duplicate both versions on both
				// platforms and warn the user.
				iterPCLastStateFavorites = colPCLastStateFavoriteMap.find(strPCName);
				bool bDuplicateBoth = false;
				if (iterPCLastStateFavorites == colPCLastStateFavoriteMap.end())
				{
					// the record doesn't exist in the "last desktop state" file
					bDuplicateBoth = true;
//					LogAddEntry("duplicating record because it could not be found in the last desktop state file", slText, true);
				}
				else
				{
					const LinkSyncDesktopFavorite& pcLastStateFavorite = (*iterPCLastStateFavorites).second;
					if (strcmp(pcLastStateFavorite.GetUrl(), pcFavorite.GetUrl()) == 0)
					{
						// the "last desktop state" agrees with the current desktop
						// state - so the handheld wins
						LinkSyncDesktopFavorite modFavorite = hhFavorite;
						modFavorite.SetDirectory(pcFavorite.GetDirectory());
						colChangeOnPC.push_back(modFavorite);
						colNewPCLastState.push_back(modFavorite);
					}
					else if (strcmp(pcLastStateFavorite.GetUrl(), hhFavorite.GetUrl()) == 0)
					{
						// the "last desktop state" agrees with the handheld -
						// so the desktop wins
						LinkSyncDesktopFavorite modFavorite = pcFavorite;
						modFavorite.SetHHRecordId(hhFavorite.GetHHRecordId());
						colChangeOnHH.push_back(modFavorite);
						colNewPCLastState.push_back(modFavorite);
					}
					else
					{
						// the "last desktop state" agrees with neither -
						// duplicate on both platforms
						bDuplicateBoth = true;
//						LogAddEntry("duplicating record because the last desktop state file agrees with neither platform", slText, true);
//						LogAddFormattedEntry(slText, true, "hh = '%s'; pc = '%s'", hhFavorite.GetUrl(), pcFavorite.GetUrl());
					}
				}

				// if we wound up in a state indicating we need to duplicate
				// on both platforms...
				if (bDuplicateBoth)
				{
					// we can't use the same name. since using a different name
					// for the new one on each platform would cause us to be
					// out of sync, BOTH names get changed
					stlport::string strDuplicateHHName = stlport::string(pcFavorite.GetName()) + stlport::string(" LINKSYNC DUPLICATE - HANDHELD");
					stlport::string strDuplicatePCName = stlport::string(pcFavorite.GetName()) + stlport::string(" LINKSYNC DUPLICATE - DESKTOP");

					// note that since we changed the names, which is what we're
					// using as the unique key, we need to add both new versions
					// to both platforms and delete the old version from each
					// platform
					colDeleteFromPC.push_back(pcFavorite);
					colDeleteFromHH.push_back(hhFavorite);

					// add the original handheld version
					LinkSyncDesktopFavorite modHHFavorite = hhFavorite;
					modHHFavorite.SetDirectory(pcFavorite.GetDirectory());
					modHHFavorite.SetName(strDuplicateHHName.c_str());
					modHHFavorite.SetHHRecordId(0);
					colAddToPC.push_back(modHHFavorite);
					colAddToHH.push_back(modHHFavorite);

					// add the original desktop version
					LinkSyncDesktopFavorite modPCFavorite = pcFavorite;
					modPCFavorite.SetHHRecordId(0);
					modPCFavorite.SetName(strDuplicatePCName.c_str());
					colAddToHH.push_back(modPCFavorite);
					colAddToPC.push_back(modPCFavorite);

					// both versions need to go in the last state file, but make
					colNewPCLastState.push_back(modHHFavorite);
					colNewPCLastState.push_back(modPCFavorite);

					// warn the user
					LogAddFormattedEntry(slWarning, true, "Record with name '%s' modified on both platforms, duplicating on each", pcFavorite.GetName());
				}
			} // if record with same name but different URLs exists on both platforms
		} // if record with this name exists on both platforms
	} // iterate over desktop records

	// next, iterate over all handheld records and see if there's any we need to
	// add/remove from the desktop
	ulYieldCount = 0;
	for (iterHHFavorites = colHHFavoriteMap.begin(); iterHHFavorites != colHHFavoriteMap.end(); iterHHFavorites++)
	{
		if (++ulYieldCount >= 100)
		{
			ulYieldCount = 0;
			err = SyncYieldCycles(1);
			if (err != 0)
			{
				LogAddFormattedEntry(slWarning, true, "Could not yeild cycles during sync while comparing records from handheld to pc, err = %d", err);
			}
		}

		const stlport::string& strHHName = (*iterHHFavorites).first;
		const LinkSyncDesktopFavorite& hhFavorite = (*iterHHFavorites).second;

		// does it exist on the desktop?
		iterPCFavorites = colPCFavoriteMap.find(strHHName);
		if (iterPCFavorites == colPCFavoriteMap.end())
		{
			// it does not exist on the desktop
			// now the question is - was it deleted from the desktop,
			// meaning we need to also delete it on the handheld? or was
			// it newly added on the handheld, meaning we need to add it
			// to the desktop?
			// the answer can be found by looking at the "last desktop state"
			// collection: if this record also exists there (by name, anyway),
			// then we know that it used to be on the desktop, but since the
			// last sync it was removed from the desktop. so therefore we need
			// to remove it from the handheld as well.
			// if this record does not exist in the "last desktop state" file,
			// then we know that it was newly added to the handheld and
			// therefore should also be added to the desktop.
			iterPCLastStateFavorites = colPCLastStateFavoriteMap.find(strHHName);
			if (iterPCLastStateFavorites == colPCLastStateFavoriteMap.end())
			{
				// no, it did not exist on the desktop as of the last sync,
				// so therefore it needs to be added to the desktop
				LinkSyncDesktopFavorite modFavorite = hhFavorite;
				modFavorite.SetDirectory(LinkSyncDesktopFavorite::GetUserFavoritesRootDirectory());
				colAddToPC.push_back(modFavorite);
				colNewPCLastState.push_back(modFavorite);
			}
			else
			{
				// yes, it did exist on the desktop as of the last sync,
				// so therefore it needs to be deleted from the handheld
				colDeleteFromHH.push_back(hhFavorite);
			}
		} // if this record does not exist on the desktop
		// note that there is no need to review the case where it exists
		// on both platforms - we already handled that while iterating over
		// the desktop items
	} // iterate over handheld records
}


void UpdateDesktopLastStateFile(CSyncProperties& rProps, const LinkSyncDesktopFavorite::FavoriteList& colPCFavorites)
{
	LinkSyncDesktopLastState lastState(rProps.m_PathName, colPCFavorites);
	lastState.WriteToFile();
}

