/******************************************************************************

  $Header: f:\\RCS\\F\\devel\\pilot\\CodeWarrior\\LinkSync\\Src\\LinkSyncMain.cpp,v 1.9 2002-08-10 20:33:40-04 eshayne Exp eshayne $
  $Revision: 1.9 $
 

   $Log: LinkSyncMain.cpp,v $
   Revision 1.9  2002-08-10 20:33:40-04  eshayne
   changed name to Link Synchronizer

   Revision 1.8  2002-08-10 18:10:11-04  eshayne
   changed (visible) name to Link Synchronizer

   Revision 1.7  2002-07-21 10:23:39-04  eshayne
   added comment and license blocks

   Revision 1.6  2002-07-21 10:18:17-04  eshayne
   added Warrany Info screen

   Revision 1.5  2002-07-20 20:56:58-04  eshayne
   added validation requiring that each record have a name and a url

   Revision 1.4  2002-04-06 21:11:54-05  eshayne
   now does edit and delete

   Revision 1.3  2002-04-03 22:40:57-05  eshayne
   edit and delete now prefill name and url fields

   Revision 1.2  2002-04-03 19:50:22-05  eshayne
   started working on edit

   Revision 1.1  2002-03-31 17:37:18-05  eshayne
   handles adds, but no auto-update, no edit, no sync, no delete...

	----------------------------------------
    Link Synchronizer
    Copyright (C)2002 Ethan Shayne

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include <PalmOS.h>

/* Palm OS Glue headers */
#include <CharAttr.h>		/* needed by TxtGlue.h */
#include <PalmOSGlue.h>		/* includes all other glue headers */

#include "LinkSync.h"
#include "LinkSyncRsc.h"

/*********************************************************************
 * Entry Points
 *********************************************************************/

/*********************************************************************
 * Global variables
 *********************************************************************/

/* g_prefs
 * cache for application preferences during program execution */
LinkSyncPreferenceType g_prefs;
DmOpenRef g_dbRef = 0;

/*********************************************************************
 * Internal Constants
 *********************************************************************/

/* Define the minimum OS version we support */
#define ourMinVersion    sysMakeROMVersion(3,0,0,sysROMStageDevelopment,0)
#define kPalmOS20Version sysMakeROMVersion(2,0,0,sysROMStageDevelopment,0)

/*********************************************************************
 * Internal Functions
 *********************************************************************/


static Err OpenDB()
{
	Err err = errNone;

	// only open the database handle if we haven't done this already...
	if (!g_dbRef) {
		g_dbRef = DmOpenDatabaseByTypeCreator('DATA', 'LSnc', dmModeReadWrite);
		if (!g_dbRef)
		{
			err = DmGetLastErr();
			// if the database doesn't exist, create it
			if (err == dmErrCantFind)
			{
				err = DmCreateDatabase(0, "Database-LSnc", 'LSnc', 'DATA', false);
				if (err)
					return err;
				
				g_dbRef = DmOpenDatabaseByTypeCreator('DATA', 'LSnc', dmModeReadWrite);
				if (!g_dbRef)
					err = DmGetLastErr();
			}
		}
	}
	
	return err;
}


static Err CloseDB()
{
	if (g_dbRef)
		return (DmCloseDatabase(g_dbRef));
	else
		return errNone;
}


// draw item # itemNum in the list on the main form
static void DrawOneString(Int16 itemNum, RectangleType* bounds, Char** itemsText)
{
	#pragma unused(itemsText)
	
	// find entry # itemNum in the database - that's the item to draw
	MemHandle record = DmQueryRecord(g_dbRef, itemNum);
	if (record)
	{
		MemPtr pEntry = MemHandleLock(record);
		// although the database entry contains both the name and the URL, the name is null terminated
		// and comes first - so pszUrlName will reflect just the name
		Char* pszUrlName = (Char*)pEntry;
		WinDrawChars(pszUrlName, StrLen(pszUrlName), bounds->topLeft.x, bounds->topLeft.y);
		MemHandleUnlock(record);
	}
}


static void MainFormInit(FormType *frmP)
{
	// open the database
	OpenDB();

	// get the current number of database entries
	UInt16 recordCount = DmNumRecordsInCategory(g_dbRef, dmAllCategories);

	ListPtr list;
	list = (ListPtr) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MainUrlsList));

	LstSetDrawFunction(list, DrawOneString);
	LstSetListChoices(list, NULL, recordCount);
}


/*
 * FUNCTION: MainFormDoCommand
 *
 * DESCRIPTION: This routine performs the menu command specified.
 *
 * PARAMETERS:
 *
 * command
 *     menu item id
 */

static Boolean MainFormDoCommand(UInt16 command)
{
    Boolean handled = false;
    FormType * frmP;

    switch (command)
    {
    case OptionsAboutLinkSynchronizer:

        /* Clear the menu status from the display */
        MenuEraseStatus(0);

        /* Display the About Box. */
        frmP = FrmInitForm (AboutForm);
        FrmDoDialog (frmP);                    
        FrmDeleteForm (frmP);

        handled = true;
        break;
        
    case OptionsWarrantyInfo:
        /* Clear the menu status from the display */
        MenuEraseStatus(0);

        /* Display the Warrany Info screen */
        frmP = FrmInitForm (WarrantyForm);
        FrmDoDialog (frmP);                    
        FrmDeleteForm (frmP);

        handled = true;
        break;
    
    }
    
    return handled;
}


static void MainFormClose()
{
	CloseDB();
}


static bool ValidateEntry(FormPtr frmP)
{
	// get handles to the name and URL fields' text
	FieldPtr pNameField = (FieldPtr) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, AddUrlUrlNameField));
	UInt32 uiNameLength = FldGetTextLength(pNameField);
	Char* pszNameText = FldGetTextPtr(pNameField);

	FieldPtr pUrlField = (FieldPtr) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, AddUrlUrlUrlField));
	UInt32 uiUrlLength = FldGetTextLength(pUrlField);
	Char* pszUrlText = FldGetTextPtr(pUrlField);
	
	if (uiNameLength == 0)
	{
		FrmCustomAlert(InvalidValuesAlert, "Name", NULL, NULL);
		return false;
	}
	
	if (uiUrlLength == 0)
	{
		FrmCustomAlert(InvalidValuesAlert, "URL", NULL, NULL);
		return false;
	}
	
	return true;
}


static void SaveEntry(FormPtr frmP, const AddFormMode eMode, const Int16 iSelection)
{
	// if this is a delete, things are much simpler...
	if (eMode == DELETE)
	{
		DmDeleteRecord(g_dbRef, iSelection);
		// move the "deleted" record to the end of the database so as not to interfere
		// with the rest of the list
		DmMoveRecord(g_dbRef, iSelection, DmNumRecords(g_dbRef));
	}
	else
	{
		// get handles to the name and URL fields' text
		FieldPtr pNameField = (FieldPtr) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, AddUrlUrlNameField));
		UInt32 uiNameLength = FldGetTextLength(pNameField);
		Char* pszNameText = FldGetTextPtr(pNameField);
	
		FieldPtr pUrlField = (FieldPtr) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, AddUrlUrlUrlField));
		UInt32 uiUrlLength = FldGetTextLength(pUrlField);
		Char* pszUrlText = FldGetTextPtr(pUrlField);
	
		MemHandle hRecord = NULL;
		UInt16 uiRecordIndex = 0;
		UInt32 uiRecordSize = uiNameLength + uiUrlLength + 2;
		switch (eMode)
		{
			// if this is an add, then create a new database entry
			case ADD:
				uiRecordIndex = 0;
				hRecord = DmNewRecord(g_dbRef, &uiRecordIndex, uiRecordSize);
				break;
			
			// otherwise, we need a handle to the existing database entry	
			default:
				uiRecordIndex = iSelection;
				hRecord = DmQueryRecord(g_dbRef, uiRecordIndex);
				MemHandleResize(hRecord, uiRecordSize);
				break;
		}
	
		// lock the record and get a pointer to its contents
		MemPtr pRecordData = MemHandleLock(hRecord);
	
		// copy the data into the record
		DmStrCopy(pRecordData, 0, pszNameText);
		DmStrCopy(pRecordData, uiNameLength + 1, pszUrlText);
	
		// cleanup
		MemHandleUnlock(hRecord);
		DmReleaseRecord(g_dbRef, uiRecordIndex, true);
	}
}


static Err SetFieldText(FieldPtr pField, const Char* pszText)
{
	MemHandle hFieldText = FldGetTextHandle(pField);
	// if there's already a text handle, use it (grow it first, though)
	if (hFieldText)
	{
		FldSetTextHandle(pField, NULL);
		Err err = MemHandleResize(hFieldText, StrLen(pszText) + 1);
		if (err != errNone)
		{
			// if it didn't work, can't do any more
			FldSetTextHandle(pField, hFieldText);
			return err;
		}
	}
	// if there wasn't already a text handle, make one and associate it
	else
	{
		hFieldText = MemHandleNew(StrLen(pszText) + 1);
		if (!hFieldText)
			return memErrNotEnoughSpace;
	}
	
	// copy the name into the name field's text handle
	StrCopy((Char*)MemHandleLock(hFieldText), pszText);
	MemHandleUnlock(hFieldText);
		
	// put the text handle back with the field
	FldSetTextHandle(pField, hFieldText);
			
	// redraw the name field now that we've changed it
	FldDrawField(pField);
	
	return errNone;
}


static void LoadAddUrlForm(FormPtr pForm, const AddFormMode eMode, const Int16 iSelection)
{
	// if this is an edit or delete, grab the name and URL from the database
	if (eMode == EDIT || eMode == DELETE)
	{
		OpenDB();
		MemHandle record = DmQueryRecord(g_dbRef, iSelection);
		if (record)
		{
			MemPtr pEntry = MemHandleLock(record);
			Char* pszUrlName = (Char*)pEntry;
			Char* pszUrlUrl = (Char*)pEntry + StrLen(pszUrlName) + 1;
			
			// prefill the name field
			FieldPtr pNameField = (FieldPtr) FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, AddUrlUrlNameField));
			SetFieldText(pNameField, pszUrlName);
	
			// prefill the URL field
			FieldPtr pUrlField = (FieldPtr) FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, AddUrlUrlUrlField));
			SetFieldText(pUrlField, pszUrlUrl);
			
			MemHandleUnlock(record);
		}
	}
}


/*
 * FUNCTION: MainFormHandleEvent
 *
 * DESCRIPTION:
 *
 * This routine is the event handler for the "MainForm" of this 
 * application.
 *
 * PARAMETERS:
 *
 * eventP
 *     a pointer to an EventType structure
 *
 * RETURNED:
 *     true if the event was handled and should not be passed to
 *     FrmHandleEvent
 */

static Boolean MainFormHandleEvent(EventType * eventP)
{
    Boolean handled = false;
    FormType * frmP;

    switch (eventP->eType) 
        {
        case menuEvent:
            return MainFormDoCommand(eventP->data.menu.itemID);

		// if we're just opening the form or if we're updating it (someone added, deleted or edited
		// an entry) - either way, it's the same set of actions to draw/redraw the form
        case frmOpenEvent:
        case frmUpdateEvent:
            frmP = FrmGetActiveForm();
            MainFormInit(frmP);
            FrmDrawForm(frmP);
            handled = true;
            break;
            
		// if the user hit a button, any button...
        case ctlSelectEvent:
        	UInt16 uiId = eventP->data.ctlSelect.controlID;
        	switch (uiId) {
				// in all three cases, we do basically the same thing - only
				// difference is how we set geMode before continuing...
        		case MainAddButton:
        		case MainEditButton:
        		case MainDeleteButton:
        			// determine which item the user selected
					ListPtr pList;
					frmP = FrmGetActiveForm();
					pList = (ListPtr) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MainUrlsList));
        			Int16 iSelection = LstGetSelection(pList);
        			
        			// figure out which button the user clicked
        			AddFormMode eMode;
        			switch (uiId)
        			{
        				case MainAddButton: eMode = ADD; break;
        				case MainEditButton: eMode = EDIT; break;
        				case MainDeleteButton: eMode = DELETE; break;
        				default: eMode = ADD; break;
        			}
        			
        			// bring up the add/edit/delete form
		        	FormPtr pAddUrlForm = FrmInitForm(AddUrlForm);
		        	FrmDrawForm(pAddUrlForm);
		        	LoadAddUrlForm(pAddUrlForm, eMode, iSelection);
		        	bool bValidated = false;
		        	UInt16 uiAddUrlButton;
		        	while (!bValidated)
		        	{
		        		uiAddUrlButton = FrmDoDialog(pAddUrlForm);
		        		if (uiAddUrlButton == AddUrlCancelButton || eMode == DELETE)
		        			bValidated = true;
		        		else
		        			bValidated = ValidateEntry(pAddUrlForm);
		        	}
		        	
		        	// if the user hit "ok," act on it
		        	if (uiAddUrlButton == AddUrlOkButton)
		        	{
	        			SaveEntry(pAddUrlForm, eMode, iSelection);
 	        			FrmUpdateForm(MainForm, 0);
		        	}
		        	
		        	// cleanup
		        	FrmDeleteForm(pAddUrlForm);
		        	
    		    	handled = true;
        			break;
        	}
        	break;
        	
        case frmCloseEvent:
        	MainFormClose();
        	handled = false;
        	break;

        default:
            break;
        }
    
    return handled;
}


/*
 * FUNCTION: AppHandleEvent
 *
 * DESCRIPTION: 
 *
 * This routine loads form resources and set the event handler for
 * the form loaded.
 *
 * PARAMETERS:
 *
 * event
 *     a pointer to an EventType structure
 *
 * RETURNED:
 *     true if the event was handled and should not be passed
 *     to a higher level handler.
 */

static Boolean AppHandleEvent(EventType * eventP)
{
    UInt16 formId;
    FormType * frmP;

    if (eventP->eType == frmLoadEvent)
    {
        /* Load the form resource. */
        formId = eventP->data.frmLoad.formID;
        frmP = FrmInitForm(formId);
        FrmSetActiveForm(frmP);

        /* Set the event handler for the form.  The handler of the
         * currently active form is called by FrmHandleEvent each
         * time is receives an event. */
        switch (formId)
        {
        case MainForm:
            FrmSetEventHandler(frmP, MainFormHandleEvent);
            break;
            
        default:
            break;

        }
        return true;
    }

    return false;
}

/*
 * FUNCTION: AppEventLoop
 *
 * DESCRIPTION: This routine is the event loop for the application.
 */

static void AppEventLoop(void)
{
    UInt16 error;
    EventType event;

    do {
        /* change timeout if you need periodic nilEvents */
        EvtGetEvent(&event, evtWaitForever);

        if (! SysHandleEvent(&event))
        {
            if (! MenuHandleEvent(0, &event, &error))
            {
                if (! AppHandleEvent(&event))
                {
                    FrmDispatchEvent(&event);
                }
            }
        }
    } while (event.eType != appStopEvent);
}

/*
 * FUNCTION: AppStart
 *
 * DESCRIPTION:  Get the current application's preferences.
 *
 * RETURNED:
 *     errNone - if nothing went wrong
 */

static Err AppStart(void)
{
    UInt16 prefsSize;

    /* Read the saved preferences / saved-state information. */
    prefsSize = sizeof(LinkSyncPreferenceType);
    if (PrefGetAppPreferences(
        appFileCreator, appPrefID, &g_prefs, &prefsSize, true) != 
        noPreferenceFound)
    {
        /* FIXME: setup g_prefs with default values */
    }
    
    return errNone;
}

/*
 * FUNCTION: AppStop
 *
 * DESCRIPTION: Save the current state of the application.
 */

static void AppStop(void)
{
    /* Write the saved preferences / saved-state information.  This
     * data will be saved during a HotSync backup. */
    PrefSetAppPreferences(
        appFileCreator, appPrefID, appPrefVersionNum, 
        &g_prefs, sizeof(LinkSyncPreferenceType), true);
        
    /* Close all the open forms. */
    FrmCloseAllForms();
}

/* all code from here to end of file should use no global variables */
#pragma warn_a5_access on

/*
 * FUNCTION: RomVersionCompatible
 *
 * DESCRIPTION: 
 *
 * This routine checks that a ROM version is meet your minimum 
 * requirement.
 *
 * PARAMETERS:
 *
 * requiredVersion
 *     minimum rom version required
 *     (see sysFtrNumROMVersion in SystemMgr.h for format)
 *
 * launchFlags
 *     flags that indicate if the application UI is initialized
 *     These flags are one of the parameters to your app's PilotMain
 *
 * RETURNED:
 *     error code or zero if ROM version is compatible
 */

static Err RomVersionCompatible(UInt32 requiredVersion, UInt16 launchFlags)
{
    UInt32 romVersion;

    /* See if we're on in minimum required version of the ROM or later. */
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    if (romVersion < requiredVersion)
    {
        if ((launchFlags & 
            (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
            (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
        {
            FrmAlert (RomIncompatibleAlert);

            /* Palm OS versions before 2.0 will continuously relaunch this
             * app unless we switch to another safe one. */
            if (romVersion < kPalmOS20Version)
            {
                AppLaunchWithCommand(
                    sysFileCDefaultApp, 
                    sysAppLaunchCmdNormalLaunch, NULL);
            }
        }

        return sysErrRomIncompatible;
    }

    return errNone;
}

/*
 * FUNCTION: LinkSyncPalmMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:
 *
 * cmd
 *     word value specifying the launch code. 
 *
 * cmdPB
 *     pointer to a structure that is associated with the launch code
 *
 * launchFlags
 *     word value providing extra information about the launch
 *
 * RETURNED:
 *     Result of launch, errNone if all went OK
 */

static UInt32 LinkSyncPalmMain(
    UInt16 cmd, 
    MemPtr /*cmdPBP*/, 
    UInt16 launchFlags)
{
    Err error;

    error = RomVersionCompatible (ourMinVersion, launchFlags);
    if (error) return (error);

    switch (cmd)
    {
    case sysAppLaunchCmdNormalLaunch:
        error = AppStart();
        if (error) 
            return error;

        /* start application by opening the main form
         * and then entering the main event loop */
        FrmGotoForm(MainForm);
        AppEventLoop();
        
        AppStop();
        break;

    default:
        break;
    }

    return errNone;
}

/*
 * FUNCTION: PilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 * 
 * PARAMETERS:
 *
 * cmd
 *     word value specifying the launch code. 
 *
 * cmdPB
 *     pointer to a structure that is associated with the launch code
 *
 * launchFlags
 *     word value providing extra information about the launch.
 *
 * RETURNED:
 *     Result of launch, errNone if all went OK
 */
 
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
    return LinkSyncPalmMain(cmd, cmdPBP, launchFlags);
}

/* turn a5 warning off to prevent it being set off by C++
 * static initializer code generation */
#pragma warn_a5_access reset
